/* OperServ functions.
 *
 * Epona (c) 2000-2001 PegSoft
 * Contact us at epona@pegsoft.net
 *
 * This program is free but copyrighted software; see the file COPYING for
 * details.
 *
 * Based on the original code of Services by Andy Church.
 *
 */

#include "services.h"
#include "pseudo.h"

/*************************************************************************/

struct clone {
    char *host;
    long time;
};

/* List of most recent users - statically initialized to zeros */
static struct clone clonelist[CLONE_DETECT_SIZE];

/* Which hosts have we warned about, and when?  This is used to keep us
 * from sending out notices over and over for clones from the same host. */
static struct clone warnings[CLONE_DETECT_SIZE];

/* List of Services administrators */
SList servadmins;
/* List of Services operators */
SList servopers;
/* AKILL, SGLINE, SQLINE and SZLINE lists */
SList akills, sglines, sqlines, szlines;

/*************************************************************************/

static void get_operserv_stats(long *nrec, long *memuse);

static int compare_adminlist_entries(SList *slist, void *item1, void *item2);
static int compare_operlist_entries(SList *slist, void *item1, void *item2);
static void free_adminlist_entry(SList *slist, void *item);
static void free_operlist_entry(SList *slist, void *item);

static int is_akill_entry_equal(SList *slist, void *item1, void *item2);
static void free_akill_entry(SList *slist, void *item);
#ifdef IRC_BAHAMUT
static int is_sgline_entry_equal(SList *slist, void *item1, void *item2);
static void free_sgline_entry(SList *slist, void *item);
#endif
static int is_sqline_entry_equal(SList *slist, void *item1, void *item2);
static void free_sqline_entry(SList *slist, void *item);
#ifdef IRC_BAHAMUT
static int is_szline_entry_equal(SList *slist, void *item1, void *item2);
static void free_szline_entry(SList *slist, void *item);
#endif

static void do_help(User *u);
static void do_global(User *u);
static void do_stats(User *u);
static void do_admin(User *u);
static void do_oper(User *u);
static void do_os_mode(User *u);
static void do_clearmodes(User *u);
static void do_os_kick(User *u);
static void do_akill(User *u);
static void do_sgline(User *u);
static void do_sqline(User *u);
static void do_szline(User *u);
static void do_set(User *u);
static void do_noop(User *u);
static void do_jupe(User *u);
static void do_raw(User *u);
static void do_update(User *u);
static void do_reload(User *u);
static void do_os_quit(User *u);
static void do_shutdown(User *u);
static void do_restart(User *u);
static void do_listignore(User *u);
static void do_killclones(User *u);
static void do_chanlist(User *u);
static void do_userlist(User *u);

#ifdef DEBUG_COMMANDS
static void send_clone_lists(User *u);
static void do_matchwild(User *u);
#endif

/*************************************************************************/

/* Options for the lists */
SListOpts akopts = { 0, NULL, &is_akill_entry_equal, &free_akill_entry };
SListOpts saopts = { SLISTF_SORT, &compare_adminlist_entries, NULL, &free_adminlist_entry };
#ifdef IRC_BAHAMUT
SListOpts sgopts = { 0, NULL, &is_sgline_entry_equal, &free_sgline_entry };
#endif
SListOpts soopts = { SLISTF_SORT, &compare_operlist_entries, NULL, &free_operlist_entry };
SListOpts sqopts = { SLISTF_SORT, NULL, &is_sqline_entry_equal, &free_sqline_entry };
#ifdef IRC_BAHAMUT
SListOpts szopts = { 0, NULL, &is_szline_entry_equal, &free_szline_entry };
#endif

/*************************************************************************/

static Command cmds[] = {
    { "HELP",       do_help,       NULL,  -1,                   -1,-1,-1,-1 },
    { "GLOBAL",     do_global,     NULL,  OPER_HELP_GLOBAL,     -1,-1,-1,-1 },
    { "STATS",      do_stats,      NULL,  OPER_HELP_STATS,      -1,-1,-1,-1 },
    { "UPTIME",     do_stats,      NULL,  OPER_HELP_STATS,      -1,-1,-1,-1 },

    /* Anyone can use the LIST option to the ADMIN and OPER commands; those
     * routines check privileges to ensure that only authorized users
     * modify the list. */
    { "ADMIN",      do_admin,      NULL,  OPER_HELP_ADMIN,      -1,-1,-1,-1 },
    { "OPER",       do_oper,       NULL,  OPER_HELP_OPER,       -1,-1,-1,-1 },
    /* Similarly, anyone can use *NEWS LIST, but *NEWS {ADD,DEL} are
     * reserved for Services admins. */
    { "LOGONNEWS",  do_logonnews,  NULL,  NEWS_HELP_LOGON,      -1,-1,-1,-1 },
    { "OPERNEWS",   do_opernews,   NULL,  NEWS_HELP_OPER,       -1,-1,-1,-1 },
    { "RANDOMNEWS", do_randomnews, NULL,  NEWS_HELP_RANDOM,     -1,-1,-1,-1 },

    /* Commands for Services opers: */
    { "MODE",       do_os_mode,    is_services_oper,
	OPER_HELP_MODE, -1,-1,-1,-1 },
    { "CLEARMODES", do_clearmodes, is_services_oper,
	OPER_HELP_CLEARMODES, -1,-1,-1,-1 },
    { "KICK",       do_os_kick,    is_services_oper,
	OPER_HELP_KICK, -1,-1,-1,-1 },
	{ "KILLCLONES", do_killclones, is_services_oper,
	OPER_HELP_KILLCLONES, -1,-1,-1, -1 },
    { "AKILL",      do_akill,      is_services_oper,
	OPER_HELP_AKILL, -1,-1,-1,-1 },
	{ "SGLINE",     do_sgline,     is_services_oper,
	OPER_HELP_SGLINE, -1,-1,-1,-1 },
	{ "SQLINE",     do_sqline,     is_services_oper,
	OPER_HELP_SQLINE, -1,-1,-1,-1 },
	{ "SZLINE",     do_szline,     is_services_oper,
	OPER_HELP_SZLINE, -1,-1,-1,-1 },

    /* Commands for Services admins: */
    { "SET",        do_set,        is_services_admin,
	OPER_HELP_SET, -1,-1,-1,-1 },
    { "SET READONLY",NULL,         NULL,  
    OPER_HELP_SET_READONLY, -1,-1,-1,-1 },
    { "SET DEBUG",  NULL,          NULL,     
    OPER_HELP_SET_DEBUG, -1,-1,-1,-1 },
    { "SET NOEXPIRE",NULL,         NULL, 
    OPER_HELP_SET_NOEXPIRE, -1,-1,-1,-1 },
    { "NOOP",       do_noop,       is_services_admin,
	OPER_HELP_NOOP, -1,-1,-1,-1 },
    { "JUPE",       do_jupe,       is_services_admin,
	OPER_HELP_JUPE, -1,-1,-1,-1 },
    { "RAW",        do_raw,        is_services_admin,
	OPER_HELP_RAW, -1,-1,-1,-1 },
    { "UPDATE",     do_update,     is_services_admin,
	OPER_HELP_UPDATE, -1,-1,-1,-1 },
	{ "RELOAD",     do_reload,     is_services_admin,
	OPER_HELP_RELOAD, -1,-1,-1,-1 },
    { "QUIT",       do_os_quit,    is_services_admin,
	OPER_HELP_QUIT, -1,-1,-1,-1 },
    { "SHUTDOWN",   do_shutdown,   is_services_admin,
	OPER_HELP_SHUTDOWN, -1,-1,-1,-1 },
    { "RESTART",    do_restart,    is_services_admin,
	OPER_HELP_RESTART, -1,-1,-1,-1 },
    { "LISTIGNORE", do_listignore, is_services_admin,
	-1,-1,-1,-1, -1 },
#ifndef STREAMLINED
    { "SESSION",    do_session,    is_services_admin,
        OPER_HELP_SESSION, -1,-1,-1, -1 },
    { "EXCEPTION",  do_exception,  is_services_admin,
        OPER_HELP_EXCEPTION, -1,-1,-1, -1 },
#endif
    { "CHANLIST",   do_chanlist,   is_services_admin,
	OPER_HELP_CHANLIST,   -1,-1,-1,-1 },
	{ "USERLIST",   do_userlist,   is_services_admin,
	OPER_HELP_USERLIST,   -1,-1,-1,-1 },
	{ "CACHE",      do_cache,      is_services_admin,
	OPER_HELP_CACHE,      -1,-1,-1,-1 },

    /* Commands for Services root: */

#ifdef DEBUG_COMMANDS
    { "LISTTIMERS", send_timeout_list,  is_services_root, -1,-1,-1,-1,-1 },
    { "MATCHWILD",  do_matchwild,       is_services_root, -1,-1,-1,-1,-1 },
    { "LISTCLONES", send_clone_lists,   is_services_root, -1,-1,-1,-1,-1 },
#endif

    /* Fencepost: */
    { NULL }
};

/*************************************************************************/
/*************************************************************************/

/* OperServ initialization. */

void os_init(void)
{
    Command *cmd;

    cmd = lookup_cmd(cmds, "GLOBAL");
    if (cmd)
	cmd->help_param1 = s_GlobalNoticer;
    cmd = lookup_cmd(cmds, "ADMIN");
    if (cmd)
	cmd->help_param1 = s_NickServ;
    cmd = lookup_cmd(cmds, "OPER");
    if (cmd)
	cmd->help_param1 = s_NickServ;
	
	/* Initialization of the lists */
	slist_init(&servadmins);
	servadmins.opts = &saopts;
	slist_init(&servopers);
	servopers.opts = &soopts;
	
	slist_init(&akills);
	akills.opts = &akopts;
	slist_init(&sglines);
#ifdef IRC_BAHAMUT
	sglines.opts = &sgopts;
#endif
	slist_init(&sqlines);
	sqlines.opts = &sqopts;
	slist_init(&szlines);
#ifdef IRC_BAHAMUT
	szlines.opts = &szopts;
#endif
}

/*************************************************************************/

/* Main OperServ routine. */

void operserv(User *u, char *buf)
{
    char *cmd;
    char *s;

    log("%s: %s: %s", s_OperServ, u->nick, buf);

    cmd = strtok(buf, " ");
    if (!cmd) {
		return;
    } else if (stricmp(cmd, "\1PING") == 0) {
		if (!(s = strtok(NULL, "")))
	    	s = "\1";
		notice(s_OperServ, u->nick, "\1PING %s", s);
    } else {
		run_cmd(s_OperServ, u, cmds, cmd);
    }
}

static void get_operserv_stats(long *nrec, long *memuse)
{
	int i;
	long mem = 0, count = 0, mem2 = 0, count2 = 0;
	Akill *ak;
	SXLine *sx;
	
	if (CheckClones) {
	    mem = sizeof(struct clone) * CLONE_DETECT_SIZE * 2;
	    for (i = 0; i < CLONE_DETECT_SIZE; i++) {
		if (clonelist[i].host) {
		    count++;
		    mem += strlen(clonelist[i].host)+1;
		}
		if (warnings[i].host) {
		    count++;
		    mem += strlen(warnings[i].host)+1;
		}
	    }
	}
	
	count += akills.count;
	mem += akills.capacity * sizeof(void);
	mem += akills.count * sizeof(Akill);
	
	for (i = 0; i < akills.count; i++) {
		ak = akills.list[i];
		mem += strlen(ak->user)+1;
		mem += strlen(ak->host)+1;
		mem += strlen(ak->by)+1;
		mem += strlen(ak->reason)+1;		
	}

#ifdef IRC_BAHAMUT	

	count += sglines.count;
	mem += sglines.capacity * sizeof(void);
	mem += sglines.count * sizeof(SXLine);
	
	for (i = 0; i < sglines.count; i++) {
		sx = sglines.list[i];
		mem += strlen(sx->mask)+1;
		mem += strlen(sx->by)+1;
		mem += strlen(sx->reason)+1;		
	}
	
#endif
	
	count += sqlines.count;
	mem += sqlines.capacity * sizeof(void);
	mem += sqlines.count * sizeof(SXLine);
	
	for (i = 0; i < sqlines.count; i++) {
		sx = sqlines.list[i];
		mem += strlen(sx->mask)+1;
		mem += strlen(sx->by)+1;
		mem += strlen(sx->reason)+1;		
	}
	
#ifdef IRC_BAHAMUT

	count += szlines.count;
	mem += szlines.capacity * sizeof(void);
	mem += szlines.count * sizeof(SXLine);
	
	for (i = 0; i < szlines.count; i++) {
		sx = szlines.list[i];
		mem += strlen(sx->mask)+1;
		mem += strlen(sx->by)+1;
		mem += strlen(sx->reason)+1;		
	}
	
#endif
	
	get_news_stats(&count2, &mem2);
	count += count2;
	mem += mem2;
	get_exception_stats(&count2, &mem2);
	count += count2;
	mem += mem2;
	
	*nrec = count;
	*memuse = mem;
}

/*************************************************************************/
/**************************** Privilege checks ***************************/
/*************************************************************************/

/* Load old AKILL data. */

#define SAFE(x) do {					\
    if ((x) < 0) {					\
	if (!forceload)					\
	    fatal("Read error on %s", AutokillDBName);	\
	break;						\
    }							\
} while (0)

static void load_old_akill(void)
{
    dbFILE *f;
    int i, j;
    int16 tmp16;
    int32 tmp32;
    char buf[NICKMAX], mask2[BUFSIZE], *mask, *s;
    Akill *ak, *entry;

    if (!(f = open_db("AKILL", AutokillDBName ? AutokillDBName : "akill.db", "r", 9)))
	return;

    get_file_version(f);

    read_int16(&tmp16, f);
    slist_setcapacity(&akills, tmp16);

	for (j = 0; j < akills.capacity; j++) {
		ak = scalloc(sizeof(Akill), 1);
		
	    SAFE(read_string(&mask, f));
	    s = strchr(mask, '@');
	    *s = 0; s++;
	    ak->user = sstrdup(mask);
	    ak->host = sstrdup(s);
	    SAFE(read_string(&ak->reason, f));
	    SAFE(read_buffer(buf, f));
	    if (!*buf) ak->by = sstrdup("<unknown>");
	    else ak->by = sstrdup(buf);
	    SAFE(read_int32(&tmp32, f));
	    ak->seton = tmp32 ? tmp32 : time(NULL);
	    SAFE(read_int32(&tmp32, f));
	    ak->expires = tmp32;
	    
	    /* Sanity checks *sigh* */
	    
	   	/* No nicknames allowed! */
	    if (strchr(ak->user, '!')) {
	    	s_rakill(ak->user, ak->host);
	    	free(ak);
	    	continue;
	    }
	    
	    snprintf(mask2, sizeof(mask2), "%s@%s", ak->user, ak->host);
	    
	    /* Is the mask already in the AKILL list? */
		if (slist_indexof(&akills, mask2) != -1) {
			free(ak);
			continue;
		}
			
		/* Checks whether there is an AKILL that already covers
	 	 * the one we want to add, and whether there are AKILLs
	 	 * that would be covered by this one. Expiry time
	 	 * does *also* matter.
	 	 */
		
		if (akills.count > 0) {
			 
			for (i = akills.count - 1; i >= 0; i--) {
		
				char amask[BUFSIZE];
				
				entry = akills.list[i];
				
				if (!entry)
					continue;
				
				snprintf(amask, sizeof(amask), "%s@%s", entry->user, entry->host);
				
				if (match_wild_nocase(amask, mask2) && (entry->expires >= ak->expires || entry->expires == 0)) {
					s_rakill(ak->user, ak->host);
					free(ak);
					ak = NULL;
					break;	
				}
				
				if (match_wild_nocase(mask2, amask) && (entry->expires <= ak->expires || ak->expires == 0))
					slist_delete(&akills, i);
			}
	
		}
	
		if (ak) 
			slist_add(&akills, ak);
	}
	
    close_db(f);
}

#undef SAFE

/* Load OperServ data. */

#define SAFE(x) do {					\
    if ((x) < 0) {					\
	if (!forceload)					\
	    fatal("Read error on %s", OperDBName);	\
	failed = 1;					\
	break;						\
    }							\
} while (0)

void load_os_dbase(void)
{
    dbFILE *f;
    int16 i, n, ver, c;
    HostCache *hc, **hclast, *hcprev;
    int16 tmp16;
    int32 tmp32;
    char *s;
    int failed = 0;

    if (!(f = open_db(s_OperServ, OperDBName, "r", OPER_VERSION)))
	return;
    
	ver = get_file_version(f);
      
   	if (ver <= 9) {
   		NickAlias *na;
   		
		SAFE(read_int16(&n, f));
		for (i = 0; i < n && !failed; i++) {
	    	SAFE(read_string(&s, f));
	    	if (s) {
	    		na = findnick(s);
	    		if (na) {
	    			na->nc->flags |= NI_SERVICES_ADMIN;
	    			if (slist_indexof(&servadmins, na) == -1)
	    			slist_add(&servadmins, na);
	    		}
	    		free(s);
	    	}
		}
		if (!failed)
	    	SAFE(read_int16(&n, f));
		for (i = 0; i < n && !failed; i++) {
	    	SAFE(read_string(&s, f));
			if (s) {
	    		na = findnick(s);
	    		if (na) {
	    			na->nc->flags |= NI_SERVICES_OPER;
	    			if (slist_indexof(&servopers, na) == -1)
	    			slist_add(&servopers, na);
	    		}
	    		free(s);
	    	}
		}
	}
		
	if (ver >= 7) {
	    int32 tmp32;
	    SAFE(read_int32(&maxusercnt, f));
	    SAFE(read_int32(&tmp32, f));
	    maxusertime = tmp32;
	}
	
	if (ver <= 10)
		load_old_akill();
	else {
		Akill *ak;
		
		read_int16(&tmp16, f);
    	slist_setcapacity(&akills, tmp16);
    	
    	for (i = 0; i < akills.capacity; i++)
    	{
    		ak = scalloc(sizeof(Akill), 1);
    		
    		SAFE(read_string(&ak->user, f));
    		SAFE(read_string(&ak->host, f));
    		SAFE(read_string(&ak->by, f));
    		SAFE(read_string(&ak->reason, f));
    		SAFE(read_int32(&tmp32, f));
    		ak->seton = tmp32;
    		SAFE(read_int32(&tmp32, f));
    		ak->expires = tmp32;
    		
    		slist_add(&akills, ak);
    	} 
	}
	
	if (ver >= 11) {
		SXLine *sx;
		
		read_int16(&tmp16, f);
    	slist_setcapacity(&sglines, tmp16);
    	
    	for (i = 0; i < sglines.capacity; i++)
    	{
    		sx = scalloc(sizeof(SXLine), 1);
    		
    		SAFE(read_string(&sx->mask, f));
    		SAFE(read_string(&sx->by, f));
    		SAFE(read_string(&sx->reason, f));
    		SAFE(read_int32(&tmp32, f));
    		sx->seton = tmp32;
    		SAFE(read_int32(&tmp32, f));
    		sx->expires = tmp32;
    		
    		slist_add(&sglines, sx);
    	}
    	
    	if (ver >= 13) {
    		read_int16(&tmp16, f);
    		slist_setcapacity(&sqlines, tmp16);
    	
    		for (i = 0; i < sqlines.capacity; i++) {
    			sx = scalloc(sizeof(SXLine), 1);
    		
    			SAFE(read_string(&sx->mask, f));
    			SAFE(read_string(&sx->by, f));
    			SAFE(read_string(&sx->reason, f));
    			SAFE(read_int32(&tmp32, f));
    			sx->seton = tmp32;
    			SAFE(read_int32(&tmp32, f));
    			sx->expires = tmp32;
    		
    			slist_add(&sqlines, sx);
    		}
    	}
    	
    	read_int16(&tmp16, f);
    	slist_setcapacity(&szlines, tmp16);
    	
    	for (i = 0; i < szlines.capacity; i++)
    	{
    		sx = scalloc(sizeof(SXLine), 1);
    		
    		SAFE(read_string(&sx->mask, f));
    		SAFE(read_string(&sx->by, f));
    		SAFE(read_string(&sx->reason, f));
    		SAFE(read_int32(&tmp32, f));
    		sx->seton = tmp32;
    		SAFE(read_int32(&tmp32, f));
    		sx->expires = tmp32;
    		
    		slist_add(&szlines, sx);
    	}
	}
	
	if (ver >= 12) {
		for (i = 0; i < 1024 && !failed; i++) {
	    	hclast = &hcache[i];
	    	hcprev = NULL;
	    		
	    	while ((c = getc_db(f)) != 0) {
	    		if (c != 1)
		    	fatal("Invalid format in %s", NickDBName);
	    		
	    		hc = scalloc(1, sizeof(HostCache));
	    		
	    		SAFE(read_string(&hc->host, f));
	    		SAFE(read_int16(&tmp16, f));
	    		hc->status = tmp16;
	    		SAFE(read_int32(&tmp32, f));
	    		hc->used = tmp32;
	    		
	    		*hclast = hc;
				hclast = &hc->next;
				hc->prev = hcprev;
				hcprev = hc;	
	    	} /* while (getc_db(f) != 0) */
	    		
	    	*hclast = NULL;
	    }	/* for (i) */
	}
	
    close_db(f);
    
}

#undef SAFE

/*************************************************************************/

/* Save OperServ data. */

#define SAFE(x) do {						\
    if ((x) < 0) {						\
	restore_db(f);						\
	log_perror("Write error on %s", OperDBName);		\
	if (time(NULL) - lastwarn > WarningTimeout) {		\
	    wallops(NULL, "Write error on %s: %s", OperDBName,	\
			strerror(errno));			\
	    lastwarn = time(NULL);				\
	}							\
	return;							\
    }								\
} while (0)

void save_os_dbase(void)
{
    int i;
    dbFILE *f;
    static time_t lastwarn = 0;
    Akill *ak;
	SXLine *sx;
	HostCache *hc;

    if (!(f = open_db(s_OperServ, OperDBName, "w", OPER_VERSION)))
	return;
    SAFE(write_int32(maxusercnt, f));
    SAFE(write_int32(maxusertime, f));
    
    SAFE(write_int16(akills.count, f));
    for (i = 0; i < akills.count; i++) {
    	ak = akills.list[i];
    	
    	SAFE(write_string(ak->user, f));
    	SAFE(write_string(ak->host, f));
    	SAFE(write_string(ak->by, f));
    	SAFE(write_string(ak->reason, f));
    	SAFE(write_int32(ak->seton, f));
    	SAFE(write_int32(ak->expires, f));
    }
    
    SAFE(write_int16(sglines.count, f));
    for (i = 0; i < sglines.count; i++) {
    	sx = sglines.list[i];
    	
    	SAFE(write_string(sx->mask, f));
    	SAFE(write_string(sx->by, f));
    	SAFE(write_string(sx->reason, f));
    	SAFE(write_int32(sx->seton, f));
    	SAFE(write_int32(sx->expires, f));
    }
    
    SAFE(write_int16(sqlines.count, f));
    for (i = 0; i < sqlines.count; i++) {
    	sx = sqlines.list[i];
    	
    	SAFE(write_string(sx->mask, f));
    	SAFE(write_string(sx->by, f));
    	SAFE(write_string(sx->reason, f));
    	SAFE(write_int32(sx->seton, f));
    	SAFE(write_int32(sx->expires, f));
    }
    
    SAFE(write_int16(szlines.count, f));
    for (i = 0; i < szlines.count; i++) {
    	sx = szlines.list[i];
    	
    	SAFE(write_string(sx->mask, f));
    	SAFE(write_string(sx->by, f));
    	SAFE(write_string(sx->reason, f));
    	SAFE(write_int32(sx->seton, f));
    	SAFE(write_int32(sx->expires, f));
    }
    
    for (i = 0; i < 1024; i++) {
		for (hc = hcache[i]; hc; hc = hc->next) {
	    	/* Don't save in-progress scans */
	    	if (hc->status < HC_NORMAL)
	    		continue;
	    	
	    	SAFE(write_int8(1, f));
	    	
	   		SAFE(write_string(hc->host, f));
	   		SAFE(write_int16(hc->status, f));
	   		SAFE(write_int32(hc->used, f));
	    	
		} /* for (hc) */
		SAFE(write_int8(0, f));
    } /* for (i) */
    
    close_db(f);
}

#undef SAFE

/*************************************************************************/

/* Removes the nick structure from OperServ lists. */

void os_remove_nick(NickCore *nc)
{
	slist_remove(&servadmins, nc);
	slist_remove(&servopers, nc);
}

/*************************************************************************/

/* Does the given user have Services root privileges?
   Now enhanced. */

int is_services_root(User *u)
{
    if ((NSStrictPrivileges && !is_oper(u)) || (!skeleton && !nick_identified(u))) 
    	return 0;
    if (skeleton || (u->na->nc->flags & NI_SERVICES_ROOT))
		return 1;
    return 0;
}

/*************************************************************************/

/* Does the given user have Services admin privileges? */

int is_services_admin(User *u)
{
    if ((NSStrictPrivileges && !is_oper(u)) || (!skeleton && !nick_identified(u)))
		return 0;
    if (skeleton || (u->na->nc->flags & (NI_SERVICES_ADMIN | NI_SERVICES_ROOT)))
		return 1;
	return 0;
}

/*************************************************************************/

/* Does the given user have Services oper privileges? */

int is_services_oper(User *u)
{
    if ((NSStrictPrivileges && !is_oper(u)) || (!skeleton && !nick_identified(u)))
		return 0;
    if (skeleton || (u->na->nc->flags & (NI_SERVICES_OPER | NI_SERVICES_ADMIN | NI_SERVICES_ROOT)))
		return 1;
	return 0;
}

/*************************************************************************/

/* Is the given nick a Services admin/root nick? */

int nick_is_services_admin(NickCore *nc)
{
    if (nc->flags & (NI_SERVICES_ADMIN | NI_SERVICES_ROOT)) 
		return 1;
	
	return 0;
}

/*************************************************************************/

/* Is the given nick a Services oper/admin/root nick? */

int nick_is_services_oper(NickCore *nc)
{
    if (nc->flags & (NI_SERVICES_OPER | NI_SERVICES_ADMIN | NI_SERVICES_ROOT)) 
		return 1;
	
	return 0;
}

/*************************************************************************/
/**************************** Clone detection ****************************/
/*************************************************************************/

/* We just got a new user; does it look like a clone?  If so, send out a
 * wallops.
 */

void check_clones(User *user)
{
#ifndef STREAMLINED
    int i, clone_count;
    long last_time;

    if (!CheckClones)
	return;

    if (clonelist[0].host)
	free(clonelist[0].host);
    i = CLONE_DETECT_SIZE-1;
    memmove(clonelist, clonelist+1, sizeof(struct clone) * i);
    clonelist[i].host = sstrdup(GetHost(user));
    last_time = clonelist[i].time = time(NULL);
    clone_count = 1;
    while (--i >= 0 && clonelist[i].host) {
	if (clonelist[i].time < last_time - CloneMaxDelay)
	    break;
	if (stricmp(clonelist[i].host, GetHost(user)) == 0) {
	    ++clone_count;
	    last_time = clonelist[i].time;
	    if (clone_count >= CloneMinUsers)
		break;
	}
    }
    if (clone_count >= CloneMinUsers) {
	/* Okay, we have clones.  Check first to see if we already know
	 * about them. */
	for (i = CLONE_DETECT_SIZE-1; i >= 0 && warnings[i].host; --i) {
	    if (stricmp(warnings[i].host, GetHost(user)) == 0)
		break;
	}
	if (i < 0 || warnings[i].time < user->my_signon - CloneWarningDelay) {
	    /* Send out the warning, and note it. */
	    wallops(s_OperServ,
		"\2WARNING\2 - possible clones detected from %s", GetHost(user));
	    log("%s: possible clones detected from %s",
		s_OperServ, GetHost(user));
	    i = CLONE_DETECT_SIZE-1;
	    if (warnings[0].host)
		free(warnings[0].host);
	    memmove(warnings, warnings+1, sizeof(struct clone) * i);
	    warnings[i].host = sstrdup(GetHost(user));
	    warnings[i].time = clonelist[i].time;
	    if (KillClones)
		kill_user(s_OperServ, user->nick, "Clone kill");
	}
    }
#endif	/* !STREAMLINED */
}

/*************************************************************************/

#ifdef DEBUG_COMMANDS

/* Send clone arrays to given nick. */

static void send_clone_lists(User *u)
{
    int i;

    if (!CheckClones) {
	notice(s_OperServ, u->nick, "CheckClones not enabled.");
	return;
    }

    notice(s_OperServ, u->nick, "clonelist[]");
    for (i = 0; i < CLONE_DETECT_SIZE; i++) {
	if (clonelist[i].host)
	    notice(s_OperServ, u->nick, "    %10ld  %s", clonelist[i].time, clonelist[i].host ? clonelist[i].host : "(null)");
    }
    notice(s_OperServ, u->nick, "warnings[]");
    for (i = 0; i < CLONE_DETECT_SIZE; i++) {
	if (clonelist[i].host)
	    notice(s_OperServ, u->nick, "    %10ld  %s", warnings[i].time, warnings[i].host ? warnings[i].host : "(null)");
    }
}

#endif	/* DEBUG_COMMANDS */

/*************************************************************************/
/*********************** OperServ command functions **********************/
/*************************************************************************/

/* HELP command. */

static void do_help(User *u)
{
    const char *cmd = strtok(NULL, "");

    if (!cmd) {
	notice_help(s_OperServ, u, OPER_HELP);
    } else {
	help_cmd(s_OperServ, u, cmds, cmd);
    }
}

/*************************************************************************/

/* Global notice sending via GlobalNoticer. */

static void do_global(User *u)
{
    char *msg = strtok(NULL, "");

    if (!msg) {
		syntax_error(s_OperServ, u, "GLOBAL", OPER_GLOBAL_SYNTAX);
		return;
    }
    
    if (WallOSGlobal)	
   		wallops(s_OperServ, "\2%s\2 just used GLOBAL command.", u->nick);

#ifdef IRC_SEQUANA
	notice(s_GlobalNoticer, "$*", "%s", msg);
#else
	if (NetworkDomain) {
		send_cmd(s_GlobalNoticer, "NOTICE $*.%s :%s", NetworkDomain, msg);
	} else {
		/* Go through all common top-level domains.  If you have others,
         * add them here.
         */
		notice(s_GlobalNoticer, "$*.com", "%s", msg);
		notice(s_GlobalNoticer, "$*.net", "%s", msg);
    	notice(s_GlobalNoticer, "$*.org", "%s", msg);
    	notice(s_GlobalNoticer, "$*.edu", "%s", msg);
	}
#endif
}

/*************************************************************************/

/* STATS command. */

static void do_stats(User *u)
{
    time_t uptime = time(NULL) - start_time;
    char *extra = strtok(NULL, "");
    int days = uptime/86400, hours = (uptime/3600)%24,
        mins = (uptime/60)%60, secs = uptime%60;
    struct tm *tm;
    char timebuf[64];

    if (extra && stricmp(extra, "ALL") != 0) {
	if (stricmp(extra, "AKILL") == 0) {
	    int timeout; 
	    /* AKILLs */
	    notice_lang(s_OperServ, u, OPER_STATS_AKILL_COUNT, akills.count);
	    timeout = AutokillExpiry+59;
	    if (timeout >= 172800)
		notice_lang(s_OperServ, u, OPER_STATS_AKILL_EXPIRE_DAYS, timeout/86400);
	    else if (timeout >= 86400)
		notice_lang(s_OperServ, u, OPER_STATS_AKILL_EXPIRE_DAY);
	    else if (timeout >= 7200)
		notice_lang(s_OperServ, u, OPER_STATS_AKILL_EXPIRE_HOURS, timeout/3600);
	    else if (timeout >= 3600)
		notice_lang(s_OperServ, u, OPER_STATS_AKILL_EXPIRE_HOUR);
	    else if (timeout >= 120)
		notice_lang(s_OperServ, u, OPER_STATS_AKILL_EXPIRE_MINS, timeout/60);
	    else if (timeout >= 60)
		notice_lang(s_OperServ, u, OPER_STATS_AKILL_EXPIRE_MIN);
	    else
		notice_lang(s_OperServ, u, OPER_STATS_AKILL_EXPIRE_NONE);
#ifdef IRC_BAHAMUT
		/* SGLINEs */
		notice_lang(s_OperServ, u, OPER_STATS_SGLINE_COUNT, sglines.count);
	    timeout = SGLineExpiry+59;
	    if (timeout >= 172800)
		notice_lang(s_OperServ, u, OPER_STATS_SGLINE_EXPIRE_DAYS, timeout/86400);
	    else if (timeout >= 86400)
		notice_lang(s_OperServ, u, OPER_STATS_SGLINE_EXPIRE_DAY);
	    else if (timeout >= 7200)
		notice_lang(s_OperServ, u, OPER_STATS_SGLINE_EXPIRE_HOURS, timeout/3600);
	    else if (timeout >= 3600)
		notice_lang(s_OperServ, u, OPER_STATS_SGLINE_EXPIRE_HOUR);
	    else if (timeout >= 120)
		notice_lang(s_OperServ, u, OPER_STATS_SGLINE_EXPIRE_MINS, timeout/60);
	    else if (timeout >= 60)
		notice_lang(s_OperServ, u, OPER_STATS_SGLINE_EXPIRE_MIN);
	    else
		notice_lang(s_OperServ, u, OPER_STATS_SGLINE_EXPIRE_NONE);
#endif
		/* SQLINEs */
		notice_lang(s_OperServ, u, OPER_STATS_SQLINE_COUNT, sqlines.count);
	    timeout = SQLineExpiry+59;
	    if (timeout >= 172800)
		notice_lang(s_OperServ, u, OPER_STATS_SQLINE_EXPIRE_DAYS, timeout/86400);
	    else if (timeout >= 86400)
		notice_lang(s_OperServ, u, OPER_STATS_SQLINE_EXPIRE_DAY);
	    else if (timeout >= 7200)
		notice_lang(s_OperServ, u, OPER_STATS_SQLINE_EXPIRE_HOURS, timeout/3600);
	    else if (timeout >= 3600)
		notice_lang(s_OperServ, u, OPER_STATS_SQLINE_EXPIRE_HOUR);
	    else if (timeout >= 120)
		notice_lang(s_OperServ, u, OPER_STATS_SQLINE_EXPIRE_MINS, timeout/60);
	    else if (timeout >= 60)
		notice_lang(s_OperServ, u, OPER_STATS_SQLINE_EXPIRE_MIN);
	    else
		notice_lang(s_OperServ, u, OPER_STATS_SQLINE_EXPIRE_NONE);
#ifdef IRC_BAHAMUT
		/* SZLINEs */
		notice_lang(s_OperServ, u, OPER_STATS_SZLINE_COUNT, szlines.count);
	    timeout = SZLineExpiry+59;
	    if (timeout >= 172800)
		notice_lang(s_OperServ, u, OPER_STATS_SZLINE_EXPIRE_DAYS, timeout/86400);
	    else if (timeout >= 86400)
		notice_lang(s_OperServ, u, OPER_STATS_SZLINE_EXPIRE_DAY);
	    else if (timeout >= 7200)
		notice_lang(s_OperServ, u, OPER_STATS_SZLINE_EXPIRE_HOURS, timeout/3600);
	    else if (timeout >= 3600)
		notice_lang(s_OperServ, u, OPER_STATS_SZLINE_EXPIRE_HOUR);
	    else if (timeout >= 120)
		notice_lang(s_OperServ, u, OPER_STATS_SZLINE_EXPIRE_MINS, timeout/60);
	    else if (timeout >= 60)
		notice_lang(s_OperServ, u, OPER_STATS_SZLINE_EXPIRE_MIN);
	    else
		notice_lang(s_OperServ, u, OPER_STATS_SZLINE_EXPIRE_NONE);
#endif
	    return;
	} else if (!stricmp(extra, "RESET")) {
		if (is_services_admin(u)) {
			maxusercnt = usercnt;
			notice_lang(s_OperServ, u, OPER_STATS_RESET);
		} else {
			notice_lang(s_OperServ, u, PERMISSION_DENIED);
		}
		return;
	} else {
	    notice_lang(s_OperServ, u, OPER_STATS_UNKNOWN_OPTION,
			extra);
	}
    }

    notice_lang(s_OperServ, u, OPER_STATS_CURRENT_USERS, usercnt, opcnt);
    tm = localtime(&maxusertime);
    strftime_lang(timebuf, sizeof(timebuf), u, STRFTIME_DATE_TIME_FORMAT, tm);
    notice_lang(s_OperServ, u, OPER_STATS_MAX_USERS, maxusercnt, timebuf);
    if (days > 1) {
	notice_lang(s_OperServ, u, OPER_STATS_UPTIME_DHMS,
		days, hours, mins, secs);
    } else if (days == 1) {
	notice_lang(s_OperServ, u, OPER_STATS_UPTIME_1DHMS,
		days, hours, mins, secs);
    } else {
	if (hours > 1) {
	    if (mins != 1) {
		if (secs != 1) {
		    notice_lang(s_OperServ, u, OPER_STATS_UPTIME_HMS,
				hours, mins, secs);
		} else {
		    notice_lang(s_OperServ, u, OPER_STATS_UPTIME_HM1S,
				hours, mins, secs);
		}
	    } else {
		if (secs != 1) {
		    notice_lang(s_OperServ, u, OPER_STATS_UPTIME_H1MS,
				hours, mins, secs);
		} else {
		    notice_lang(s_OperServ, u, OPER_STATS_UPTIME_H1M1S,
				hours, mins, secs);
		}
	    }
	} else if (hours == 1) {
	    if (mins != 1) {
		if (secs != 1) {
		    notice_lang(s_OperServ, u, OPER_STATS_UPTIME_1HMS,
				hours, mins, secs);
		} else {
		    notice_lang(s_OperServ, u, OPER_STATS_UPTIME_1HM1S,
				hours, mins, secs);
		}
	    } else {
		if (secs != 1) {
		    notice_lang(s_OperServ, u, OPER_STATS_UPTIME_1H1MS,
				hours, mins, secs);
		} else {
		    notice_lang(s_OperServ, u, OPER_STATS_UPTIME_1H1M1S,
				hours, mins, secs);
		}
	    }
	} else {
	    if (mins != 1) {
		if (secs != 1) {
		    notice_lang(s_OperServ, u, OPER_STATS_UPTIME_MS,
				mins, secs);
		} else {
		    notice_lang(s_OperServ, u, OPER_STATS_UPTIME_M1S,
				mins, secs);
		}
	    } else {
		if (secs != 1) {
		    notice_lang(s_OperServ, u, OPER_STATS_UPTIME_1MS,
				mins, secs);
		} else {
		    notice_lang(s_OperServ, u, OPER_STATS_UPTIME_1M1S,
				mins, secs);
		}
	    }
	}
    }

    if (extra && stricmp(extra, "ALL") == 0 && is_services_admin(u)) {
		long count, mem;

		notice_lang(s_OperServ, u, OPER_STATS_BYTES_READ, total_read / 1024);
		notice_lang(s_OperServ, u, OPER_STATS_BYTES_WRITTEN, total_written / 1024);

		get_user_stats(&count, &mem);
		notice_lang(s_OperServ, u, OPER_STATS_USER_MEM, count, (mem+512) / 1024);
		get_channel_stats(&count, &mem);
		notice_lang(s_OperServ, u, OPER_STATS_CHANNEL_MEM, count, (mem+512) / 1024);
		get_core_stats(&count, &mem);
		notice_lang(s_OperServ, u, OPER_STATS_GROUPS_MEM, count, (mem+512) / 1024);
		get_aliases_stats(&count, &mem);
		notice_lang(s_OperServ, u, OPER_STATS_ALIASES_MEM, count, (mem+512) / 1024);
		get_chanserv_stats(&count, &mem);
		notice_lang(s_OperServ, u, OPER_STATS_CHANSERV_MEM, count, (mem+512) / 1024);
		get_botserv_stats(&count, &mem);
		notice_lang(s_OperServ, u, OPER_STATS_BOTSERV_MEM, count, (mem+512) / 1024);
		get_operserv_stats(&count, &mem);
		notice_lang(s_OperServ, u, OPER_STATS_OPERSERV_MEM, count, (mem+512) / 1024);
		get_session_stats(&count, &mem);
		notice_lang(s_OperServ, u, OPER_STATS_SESSIONS_MEM, count, (mem+512) / 1024);
#ifdef USE_THREADS
		if (ProxyDetect) {
			get_proxy_stats(&count, &mem);
			notice_lang(s_OperServ, u, OPER_STATS_PROXY_MEM, count, (mem+512) / 1024);	
		}
#endif
    }
}

/*************************************************************************/

/* Channel mode changing (MODE command). */

static void do_os_mode(User *u)
{
    int argc;
    char **argv;
    char *s = strtok(NULL, "");
    char *chan, *modes;
    Channel *c;

    if (!s) {
	syntax_error(s_OperServ, u, "MODE", OPER_MODE_SYNTAX);
	return;
    }
    chan = s;
    s += strcspn(s, " ");
    if (!*s) {
	syntax_error(s_OperServ, u, "MODE", OPER_MODE_SYNTAX);
	return;
    }
    *s = 0;
    modes = (s+1) + strspn(s+1, " ");
    if (!*modes) {
	syntax_error(s_OperServ, u, "MODE", OPER_MODE_SYNTAX);
	return;
    }
    if (!(c = findchan(chan))) {
	notice_lang(s_OperServ, u, CHAN_X_NOT_IN_USE, chan);
    } else if (c->bouncy_modes) {
	notice_lang(s_OperServ, u, OPER_BOUNCY_MODES_U_LINE);
	return;
    } else {
	send_cmd(s_OperServ, "MODE %s %s", chan, modes);
	if (WallOSMode)
	    wallops(s_OperServ, "%s used MODE %s on %s", u->nick, modes, chan);
	*s = ' ';
	argc = split_buf(chan, &argv, 1);
	do_cmode(s_OperServ, argc, argv);
    }
}

/*************************************************************************/

/* Clear all modes from a channel. */

static void do_clearmodes(User *u)
{
    char *s;
    int i;
    char *argv[3];
    char *chan = strtok(NULL, " ");
    Channel *c;
    int all = 0;
    int count;		/* For saving ban info */
    char **bans;	/* For saving ban info */
    struct c_userlist *cu, *next;

    if (!chan) {
	syntax_error(s_OperServ, u, "CLEARMODES", OPER_CLEARMODES_SYNTAX);
    } else if (!(c = findchan(chan))) {
	notice_lang(s_OperServ, u, CHAN_X_NOT_IN_USE, chan);
    } else if (c->bouncy_modes) {
	notice_lang(s_OperServ, u, OPER_BOUNCY_MODES_U_LINE);
	return;
    } else {
	s = strtok(NULL, " ");
	if (s) {
	    if (stricmp(s, "ALL") == 0) {
		all = 1;
	    } else {
		syntax_error(s_OperServ,u,"CLEARMODES",OPER_CLEARMODES_SYNTAX);
		return;
	    }
	}
	if (WallOSClearmodes)
	    wallops(s_OperServ, "%s used CLEARMODES%s on %s",
			u->nick, all ? " ALL" : "", chan);
	if (all) {
	    /* Clear mode +o */
	    for (cu = c->users; cu; cu = next) {
		next = cu->next;
		if (!chan_has_user_status(c, cu->user, CUS_OP))
	    	continue;
		argv[0] = sstrdup(chan);
		argv[1] = sstrdup("-o");
		argv[2] = sstrdup(cu->user->nick);
		send_cmd(s_OperServ, "MODE %s %s :%s",
			argv[0], argv[1], argv[2]);
		do_cmode(s_OperServ, 3, argv);
		free(argv[2]);
		free(argv[1]);
		free(argv[0]);
	    }

	    /* Clear mode +v */
	    for (cu = c->users; cu; cu = next) {
		next = cu->next;
		if (!chan_has_user_status(c, cu->user, CUS_VOICE))
	    	continue;
		argv[0] = sstrdup(chan);
		argv[1] = sstrdup("-v");
		argv[2] = sstrdup(cu->user->nick);
		send_cmd(s_OperServ, "MODE %s %s :%s",
			argv[0], argv[1], argv[2]);
		do_cmode(s_OperServ, 3, argv);
		free(argv[2]);
		free(argv[1]);
		free(argv[0]);
	    }
	}

	/* Clear modes */
	send_cmd(s_OperServ, "MODE %s %s :%s",
		chan, MODESTOREMOVE, c->key ? c->key : "");
	argv[0] = sstrdup(chan);
	argv[1] = sstrdup(MODESTOREMOVE);
	argv[2] = c->key ? c->key : sstrdup("");
	do_cmode(s_OperServ, 2, argv);
	free(argv[0]);
	free(argv[1]);
	free(argv[2]);
	c->key = NULL;
	c->limit = 0;

	/* Clear bans */
	count = c->bancount;
	bans = scalloc(sizeof(char *) * count, 1);
	for (i = 0; i < count; i++)
	    bans[i] = sstrdup(c->bans[i]);
	for (i = 0; i < count; i++) {
	    argv[0] = sstrdup(chan);
	    argv[1] = sstrdup("-b");
	    argv[2] = bans[i];
	    send_cmd(s_OperServ, "MODE %s %s :%s",
			argv[0], argv[1], argv[2]);
	    do_cmode(s_OperServ, 3, argv);
	    free(argv[2]);
	    free(argv[1]);
	    free(argv[0]);
	}
	free(bans);
	}

	notice_lang(s_OperServ, u, OPER_CLEARMODES_ALL_DONE, chan);
}

/*************************************************************************/

/* Kick a user from a channel (KICK command). */

static void do_os_kick(User *u)
{
    char *argv[3];
    char *chan, *nick, *s;
    Channel *c;

    chan = strtok(NULL, " ");
    nick = strtok(NULL, " ");
    s = strtok(NULL, "");
    if (!chan || !nick || !s) {
	syntax_error(s_OperServ, u, "KICK", OPER_KICK_SYNTAX);
	return;
    }
    if (!(c = findchan(chan))) {
	notice_lang(s_OperServ, u, CHAN_X_NOT_IN_USE, chan);
    } else if (c->bouncy_modes) {
	notice_lang(s_OperServ, u, OPER_BOUNCY_MODES_U_LINE);
	return;
    }
    send_cmd(s_OperServ, "KICK %s %s :%s (%s)", chan, nick, u->nick, s);
    if (WallOSKick)
	wallops(s_OperServ, "%s used KICK on %s/%s", u->nick, nick, chan);
    argv[0] = sstrdup(chan);
    argv[1] = sstrdup(nick);
    argv[2] = sstrdup(s);
    do_kick(s_OperServ, 3, argv);
    free(argv[2]);
    free(argv[1]);
    free(argv[0]);
}

/*************************************************************************/

/* Adds an AKILL to the list. Returns >= 0 on success, -1 if it fails, -2
 * if only the expiry time was changed.
 * The success result is the number of AKILLs that were deleted to successfully add one.
 */

int add_akill(User *u, char *mask, const char *by, const time_t expires, const char *reason)
{
	int deleted = 0, i;
	char *user, *mask2, *host;
	Akill *entry;
			
	/* Checks whether there is an AKILL that already covers
	 * the one we want to add, and whether there are AKILLs
	 * that would be covered by this one. The masks AND the
	 * expiry times are used to determine this, because some
	 * AKILLs may become useful when another one expires.
	 * If so, warn the user in the first case and cleanup
	 * the useless AKILLs in the second.
	 */
		
	if (akills.count > 0) {
			 
	for (i = akills.count - 1; i >= 0; i--) {
		char amask[BUFSIZE];
				
		entry = akills.list[i];
				
		if (!entry)
			continue;
				
		snprintf(amask, sizeof(amask), "%s@%s", entry->user, entry->host);
				
		if (!stricmp(amask, mask)) {
			/* We change the AKILL expiry time if its current one is less than the new.
			 * This is preferable to be sure we don't change an important AKILL
			 * accidentely.
			 */
			if (entry->expires >= expires || entry->expires == 0) {
				if (u) notice_lang(s_OperServ, u, OPER_AKILL_EXISTS, mask);
				return -1;
			} else {
				entry->expires = expires;
				if (u) notice_lang(s_OperServ, u, OPER_AKILL_CHANGED, amask);
				return -2;	
			}
		}
				
		if (match_wild_nocase(amask, mask) && (entry->expires >= expires || entry->expires == 0)) {
			if (u) notice_lang(s_OperServ, u, OPER_AKILL_ALREADY_COVERED, mask, amask);
			return -1;
		}
				
		if (match_wild_nocase(mask, amask) && (entry->expires <= expires || expires == 0)) {
			slist_delete(&akills, i);
			deleted++;
		}
	}
	
	}
			
	/* We can now check whether the list is full or not. */
	if (slist_full(&akills)) {
		if (u) notice_lang(s_OperServ, u, OPER_AKILL_REACHED_LIMIT, akills.limit);
		return -1;	
	}
			
	/* We can now (really) add the AKILL. */
	mask2 = sstrdup(mask);
	host = strchr(mask2, '@');
	if (!host) return -1;
	user = mask2;
	*host = 0; host++;
			
	entry = scalloc(sizeof(Akill), 1);
	if (!entry) return -1;
				
	entry->user = sstrdup(user);
	entry->host = sstrdup(host);
	entry->by = sstrdup(by);
	entry->reason = sstrdup(reason);
	entry->seton = time(NULL);
	entry->expires = expires;
			
	slist_add(&akills, entry);
	
	if (AkillOnAdd)
		s_akill(entry->user, entry->host, entry->by, entry->seton, entry->expires, entry->reason);
	
	free(mask2);
	
	return deleted;
}

/* Does the user match any AKILLs? */

int check_akill(const char *nick, const char *username, const char *host, const char *vhost)
{
	int i;
	Akill *ak;
	
	if (akills.count == 0)
		return 0;
	
	for (i = 0; i < akills.count; i++) {
		ak = akills.list[i];
		if (!ak) continue;
		if (match_wild_nocase(ak->user, username) && (match_wild_nocase(ak->host, host) || (vhost && match_wild_nocase(ak->host, vhost)))) {
			s_akill(ak->user, ak->host, ak->by, ak->seton, ak->expires, ak->reason);
			return 1;
		}
	}
	
	return 0;
}

/* Delete any expired autokills. */

void expire_akills(void)
{
    int i;
	time_t now = time(NULL);
	Akill *ak;

	for (i = akills.count - 1; i >= 0; i--) {
		ak = akills.list[i];
		
		if (!ak->expires || ak->expires > now)
			continue;
			
		if (WallAkillExpire)
	    	wallops(s_OperServ, "AKILL on %s@%s has expired", ak->user, ak->host);
	    slist_delete(&akills, i);
	}
}

static void free_akill_entry(SList *slist, void *item)
{
	Akill *ak = item;
	
	/* Remove the AKILLs from all the servers */
	s_rakill(ak->user, ak->host);
	
	/* Free the structure */
	free(ak->user);
	free(ak->host);
	free(ak->by);
	free(ak->reason);
	free(ak);
}

/* item1 is not an Akill pointer, but a char 
 */

static int is_akill_entry_equal(SList *slist, void *item1, void *item2)
{
	char *ak1 = item1, buf[BUFSIZE];
	Akill *ak2 = item2;	
	
	if (!ak1 || !ak2) return 0;
	
	snprintf(buf, sizeof(buf), "%s@%s", ak2->user, ak2->host);
	
	if (!stricmp(ak1, buf))
		return 1;
	else
		return 0;	
}

/* Lists an AKILL entry, prefixing it with the header if needed */

static int akill_list(int number, Akill *ak, User *u, int *sent_header)
{
	char mask[BUFSIZE];
	
	if (!ak) 
	return 0;
	
	if (!*sent_header) {
		notice_lang(s_OperServ, u, OPER_AKILL_LIST_HEADER);
		*sent_header = 1;
	}
	
	snprintf(mask, sizeof(mask), "%s@%s", ak->user, ak->host);
	notice_lang(s_OperServ, u, OPER_AKILL_LIST_FORMAT, number, mask, ak->reason);
		
	return 1; 	
}

/* Callback for enumeration purposes */

static int akill_list_callback(SList *slist, int number, void *item, va_list args)
{
	User *u = va_arg(args, User *);
	int *sent_header = va_arg(args, int *);
	
	return akill_list(number, item, u, sent_header);
}

/* Lists an AKILL entry, prefixing it with the header if needed */

static int akill_view(int number, Akill *ak, User *u, int *sent_header)
{
	char mask[BUFSIZE];
	char timebuf[32], expirebuf[256];
	struct tm tm;
	
	if (!ak) 
	return 0;
	
	if (!*sent_header) {
		notice_lang(s_OperServ, u, OPER_AKILL_VIEW_HEADER);
		*sent_header = 1;
	}
	
	snprintf(mask, sizeof(mask), "%s@%s", ak->user, ak->host);
	tm = *localtime(&ak->seton);
	strftime_lang(timebuf, sizeof(timebuf), u, STRFTIME_SHORT_DATE_FORMAT, &tm);
	expire_left(u->na, expirebuf, sizeof(expirebuf), ak->expires);
	notice_lang(s_OperServ, u, OPER_AKILL_VIEW_FORMAT, number, mask, ak->by, timebuf, expirebuf, ak->reason);
		
	return 1; 	
}

/* Callback for enumeration purposes */

static int akill_view_callback(SList *slist, int number, void *item, va_list args)
{
	User *u = va_arg(args, User *);
	int *sent_header = va_arg(args, int *);
	
	return akill_view(number, item, u, sent_header);
}

/* Manage the AKILL list. */

static void do_akill(User *u)
{
	char *cmd = strtok(NULL, " ");
	
	if (!cmd)
	cmd = "";
	
	if (!stricmp(cmd, "ADD")) {
		int deleted = 0;
		char *expiry, *mask, *reason;
		time_t expires;
		
		mask = strtok(NULL, " ");
		if (mask && *mask == '+') {
			expiry = mask;
			mask = strtok(NULL, " ");
		} else {
			expiry = NULL;	
		}
		
		expires = expiry ? dotime(expiry) : AutokillExpiry;
		/* If the expiry given does not contain a final letter, it's in days,
		 * said the doc. Ah well.
		 */
		if (expiry && isdigit(expiry[strlen(expiry)-1]))
	    		expires *= 86400;
		/* Do not allow less than a minute expiry time */
		if (expires != 0 && expires < 60) {
	    	notice_lang(s_OperServ, u, BAD_EXPIRY_TIME);
	    	return;
		} else if (expires > 0) {
	    	expires += time(NULL);
		}
		
		if (mask && (reason = strtok(NULL, ""))) {
			/* We first do some sanity check on the proposed mask. */
			if (strchr(mask, '!')) {
				notice_lang(s_OperServ, u, OPER_AKILL_NO_NICK);
				return;	
			}
			
			if (!strchr(mask, '@')) {
				notice_lang(s_OperServ, u, BAD_USERHOST_MASK);
				return;
			}
			
			if (mask && strspn(mask, "~@.*?") == strlen(mask)) {
				notice_lang(s_OperServ, u, USERHOST_MASK_TOO_WIDE, mask);
				return;
			}
			
			deleted = add_akill(u, mask, u->nick, expires, reason);
			if (deleted < 0)
				return;
			else if (deleted)
				notice_lang(s_OperServ, u, OPER_AKILL_DELETED_SEVERAL, deleted);
			notice_lang(s_OperServ, u, OPER_AKILL_ADDED, mask); 
			
			if (WallOSAkill) {
				char buf[128];
				
				if(!expires) {
					strcpy(buf, "does not expire");
				} else {
					int wall_expiry = expires-time(NULL);
					char *s = NULL;
					
					if (wall_expiry >= 86400) {
						wall_expiry /= 86400;
						s = "day";
					} else if (wall_expiry >= 3600) { 
						wall_expiry /= 3600;
						s = "hour";
					} else if (wall_expiry >= 60) {
						wall_expiry /= 60;
						s = "minute";
					}
					
					snprintf(buf, sizeof(buf), "expires in %d %s%s", wall_expiry, s, (wall_expiry == 1) ? "" : "s");
				}
				
				wallops(s_OperServ, "%s added an AKILL for %s (%s)", u->nick, mask, buf);
			}
			
			if (readonly)
				notice_lang(s_OperServ, u, READ_ONLY_MODE);
			
		} else {
			syntax_error(s_OperServ, u, "AKILL", OPER_AKILL_SYNTAX);
		}
		
	} else if (!stricmp(cmd, "DEL")) {
		
		char *mask;
		int res = 0;
		
		mask = strtok(NULL, " ");
		
		if (!mask) {
			syntax_error(s_OperServ, u, "AKILL", OPER_AKILL_SYNTAX);
			return; 
		}
		
		if (akills.count == 0) {
			notice_lang(s_OperServ, u, OPER_AKILL_LIST_EMPTY);
			return;
		}
		
		if (isdigit(*mask) && strspn(mask, "1234567890,-") == strlen(mask)) {
			/* Deleting a range */
			res = slist_delete_range(&akills, mask, NULL);
			if (res == 0) {
				notice_lang(s_OperServ, u, OPER_AKILL_NO_MATCH);
				return;
			} else if (res == 1) {
				notice_lang(s_OperServ, u, OPER_AKILL_DELETED_ONE);
			} else {
				notice_lang(s_OperServ, u, OPER_AKILL_DELETED_SEVERAL, res);
			}
		} else {
			if ((res = slist_indexof(&akills, mask)) == -1) {
				notice_lang(s_OperServ, u, OPER_AKILL_NOT_FOUND, mask);
				return;	
			}
			
			slist_delete(&akills, res);
			notice_lang(s_OperServ, u, OPER_AKILL_DELETED, mask);
		}
		
		if (readonly)
			notice_lang(s_OperServ, u, READ_ONLY_MODE);
		
	} else if (!stricmp(cmd, "LIST")) {
		char *mask;
		int res, sent_header = 0;
	
		if (akills.count == 0) {
			notice_lang(s_OperServ, u, OPER_AKILL_LIST_EMPTY);
			return;
		}
		
		mask = strtok(NULL, " ");
	
		if (!mask || (isdigit(*mask) && strspn(mask, "1234567890,-") == strlen(mask))) {
			res = slist_enum(&akills, mask, &akill_list_callback, u, &sent_header);
			if (res == 0) {
				notice_lang(s_OperServ, u, OPER_AKILL_NO_MATCH);
				return;
			}
		} else {
			int i;
			char amask[BUFSIZE];
			
			for (i = 0;	i < akills.count; i++) {
				snprintf(amask, sizeof(amask), "%s@%s", ((Akill *)akills.list[i])->user, ((Akill *)akills.list[i])->host);
				if (!stricmp(mask, amask) || match_wild_nocase(mask, amask))
					akill_list(i+1, akills.list[i], u, &sent_header);
			}	
		
			if (!sent_header)
			notice_lang(s_OperServ, u, OPER_AKILL_NO_MATCH);
		}
	} else if (!stricmp(cmd, "VIEW")) {
		char *mask;
		int res, sent_header = 0;
	
		if (akills.count == 0) {
			notice_lang(s_OperServ, u, OPER_AKILL_LIST_EMPTY);
			return;
		}
		
		mask = strtok(NULL, " ");
	
		if (!mask || (isdigit(*mask) && strspn(mask, "1234567890,-") == strlen(mask))) {
			res = slist_enum(&akills, mask, &akill_view_callback, u, &sent_header);
			if (res == 0) {
				notice_lang(s_OperServ, u, OPER_AKILL_NO_MATCH);
				return;
			}
		} else {
			int i;
			char amask[BUFSIZE];
			
			for (i = 0;	i < akills.count; i++) {
				snprintf(amask, sizeof(amask), "%s@%s", ((Akill *)akills.list[i])->user, ((Akill *)akills.list[i])->host);
				if (!stricmp(mask, amask) || match_wild_nocase(mask, amask))
					akill_view(i+1, akills.list[i], u, &sent_header);
			}	
		
			if (!sent_header)
			notice_lang(s_OperServ, u, OPER_AKILL_NO_MATCH);
		}
	} else if (!stricmp(cmd, "CLEAR")) {
		slist_clear(&akills, 1);
		notice_lang(s_OperServ, u, OPER_AKILL_CLEAR);
	} else {
		syntax_error(s_OperServ, u, "AKILL", OPER_AKILL_SYNTAX);
	}
}

/*************************************************************************/

#ifdef IRC_BAHAMUT

/* Adds an SGLINE to the list. Returns >= 0 on success, -1 if it failed, -2 if
 * only the expiry time changed.
 * The success result is the number of SGLINEs that were deleted to successfully add one.
 */

int add_sgline(User *u, char *mask, const char *by, const time_t expires, const char *reason)
{
	int deleted = 0, i;
	SXLine *entry;
	
	/* Checks whether there is an SGLINE that already covers
	 * the one we want to add, and whether there are SGLINEs
	 * that would be covered by this one.
	 * If so, warn the user in the first case and cleanup
	 * the useless SGLINEs in the second.
	 */
		
	if (sglines.count > 0) {
			 
	for (i = sglines.count - 1; i >= 0; i--) {
		entry = sglines.list[i];
				
		if (!entry)
			continue;
				
		if (!stricmp(entry->mask, mask)) {
			if (entry->expires >= expires || entry->expires == 0) {
				if (u) notice_lang(s_OperServ, u, OPER_SGLINE_EXISTS, mask);
				return -1;
			} else {
				entry->expires = expires;
				if (u) notice_lang(s_OperServ, u, OPER_SGLINE_CHANGED, entry->mask);
				return -2;	
			}	
		}
				
		if (match_wild_nocase(entry->mask, mask) && (entry->expires >= expires || entry->expires == 0)) {
			if (u) notice_lang(s_OperServ, u, OPER_SGLINE_ALREADY_COVERED, mask, entry->mask);
			return -1;
		}
				
		if (match_wild_nocase(mask, entry->mask) && (entry->expires <= expires || expires == 0)) {
			slist_delete(&sglines, i);
			deleted++;
		}
	}
	
	}
			
	/* We can now check whether the list is full or not. */
	if (slist_full(&sglines)) {
		if (u) notice_lang(s_OperServ, u, OPER_SGLINE_REACHED_LIMIT, sglines.limit);
		return -1;	
	}
			
	/* We can now (really) add the SGLINE. */
	entry = scalloc(sizeof(SXLine), 1);
	if (!entry) return -1;
	
	entry->mask = sstrdup(mask);
	entry->by = sstrdup(by);
	entry->reason = sstrdup(reason);
	entry->seton = time(NULL);
	entry->expires = expires;
			
	slist_add(&sglines, entry);
	
	s_sgline(entry->mask, entry->reason);
	
	return deleted;
}

/* Does the user match any SGLINEs? */

int check_sgline(const char *nick, const char *realname)
{
	int i;
	SXLine *sx;
	
	if (sglines.count == 0)
		return 0;
	
	for (i = 0; i < sglines.count; i++) {
		sx = sglines.list[i];
		if (!sx) continue;
		
		if (match_wild_nocase(sx->mask, realname)) {
			s_sgline(sx->mask, sx->reason);
			/* We kill nick since s_sgline can't */
			send_cmd(ServerName, "SVSKILL %s :G-Lined: %s", nick, sx->reason);  
			return 1;
		}
	}
	
	return 0;
}

/* Delete any expired SGLINEs. */

void expire_sglines(void)
{
    int i;
	time_t now = time(NULL);
	SXLine *sx;

	for (i = sglines.count - 1; i >= 0; i--) {
		sx = sglines.list[i];
		
		if (!sx->expires || sx->expires > now)
			continue;
			
		if (WallSGLineExpire)
	    	wallops(s_OperServ, "SGLINE on \2%s\2 has expired", sx->mask);
	    slist_delete(&sglines, i);
	}
}

static void free_sgline_entry(SList *slist, void *item)
{
	SXLine *sx = item;
	
	/* Remove the SGLINE from all the servers */
	s_unsgline(sx->mask);
	
	/* Free the structure */
	free(sx->mask);
	free(sx->by);
	free(sx->reason);
	free(sx);
}

/* item1 is not an SXLine pointer, but a char */

static int is_sgline_entry_equal(SList *slist, void *item1, void *item2)
{
	char *sx1 = item1;
	SXLine *sx2 = item2;	
	
	if (!sx1 || !sx2) return 0;
	
	if (!stricmp(sx1, sx2->mask))
		return 1;
	else
		return 0;	
}

/* Lists an SGLINE entry, prefixing it with the header if needed */

static int sgline_list(int number, SXLine *sx, User *u, int *sent_header)
{
	if (!sx) 
	return 0;
	
	if (!*sent_header) {
		notice_lang(s_OperServ, u, OPER_SGLINE_LIST_HEADER);
		*sent_header = 1;
	}
	
	notice_lang(s_OperServ, u, OPER_SGLINE_LIST_FORMAT, number, sx->mask, sx->reason);
		
	return 1; 	
}

/* Callback for enumeration purposes */

static int sgline_list_callback(SList *slist, int number, void *item, va_list args)
{
	User *u = va_arg(args, User *);
	int *sent_header = va_arg(args, int *);
	
	return sgline_list(number, item, u, sent_header);
}

/* Lists an SGLINE entry, prefixing it with the header if needed */

static int sgline_view(int number, SXLine *sx, User *u, int *sent_header)
{
	char timebuf[32], expirebuf[256];
	struct tm tm;
	
	if (!sx) 
	return 0;
	
	if (!*sent_header) {
		notice_lang(s_OperServ, u, OPER_SGLINE_VIEW_HEADER);
		*sent_header = 1;
	}
	
	tm = *localtime(&sx->seton);
	strftime_lang(timebuf, sizeof(timebuf), u, STRFTIME_SHORT_DATE_FORMAT, &tm);
	expire_left(u->na, expirebuf, sizeof(expirebuf), sx->expires);
	notice_lang(s_OperServ, u, OPER_SGLINE_VIEW_FORMAT, number, sx->mask, sx->by, timebuf, expirebuf, sx->reason);
		
	return 1; 	
}

/* Callback for enumeration purposes */

static int sgline_view_callback(SList *slist, int number, void *item, va_list args)
{
	User *u = va_arg(args, User *);
	int *sent_header = va_arg(args, int *);
	
	return sgline_view(number, item, u, sent_header);
}

#endif

/* Manage the SGLINE list. */

static void do_sgline(User *u)
{
#ifdef IRC_BAHAMUT
	char *cmd = strtok(NULL, " ");
	
	if (!cmd)
	cmd = "";
	
	if (!stricmp(cmd, "ADD")) {
		int deleted = 0;
		char *expiry, *mask, *reason;
		time_t expires;
		
		mask = strtok(NULL, ":");
		if (mask && *mask == '+') {
			expiry = mask;
			mask = strchr(expiry, ' ');
			if (mask) {
				*mask = 0;
				mask++;
			}
		} else {
			expiry = NULL;	
		}
		
		expires = expiry ? dotime(expiry) : SGLineExpiry;
		/* If the expiry given does not contain a final letter, it's in days,
		 * said the doc. Ah well.
		 */
		if (expiry && isdigit(expiry[strlen(expiry)-1]))
	    		expires *= 86400;
		/* Do not allow less than a minute expiry time */
		if (expires != 0 && expires < 60) {
	    	notice_lang(s_OperServ, u, BAD_EXPIRY_TIME);
	    	return;
		} else if (expires > 0) {
	    	expires += time(NULL);
		}
		
		if (mask && (reason = strtok(NULL, ""))) {
			/* We first do some sanity check on the proposed mask. */
			
			if (mask && strspn(mask, "*?") == strlen(mask)) {
				notice_lang(s_OperServ, u, USERHOST_MASK_TOO_WIDE, mask);
				return;
			}
			
			deleted = add_sgline(u, mask, u->nick, expires, reason);
			if (deleted < 0)
				return;
			else if (deleted)
				notice_lang(s_OperServ, u, OPER_SGLINE_DELETED_SEVERAL, deleted);
			notice_lang(s_OperServ, u, OPER_SGLINE_ADDED, mask); 
			
			if (WallOSSGLine) {
				char buf[128];
				
				if(!expires) {
					strcpy(buf, "does not expire");
				} else {
					int wall_expiry = expires-time(NULL);
					char *s = NULL;
					
					if (wall_expiry >= 86400) {
						wall_expiry /= 86400;
						s = "day";
					} else if (wall_expiry >= 3600) { 
						wall_expiry /= 3600;
						s = "hour";
					} else if (wall_expiry >= 60) {
						wall_expiry /= 60;
						s = "minute";
					}
					
					snprintf(buf, sizeof(buf), "expires in %d %s%s", wall_expiry, s, (wall_expiry == 1) ? "" : "s");
				}
				
				wallops(s_OperServ, "%s added an SGLINE for %s (%s)", u->nick, mask, buf);
			}
			
			if (readonly)
				notice_lang(s_OperServ, u, READ_ONLY_MODE);
			
		} else {
			syntax_error(s_OperServ, u, "SGLINE", OPER_SGLINE_SYNTAX);
		}
		
	} else if (!stricmp(cmd, "DEL")) {
		
		char *mask;
		int res = 0;
		
		mask = strtok(NULL, "");
		
		if (!mask) {
			syntax_error(s_OperServ, u, "SGLINE", OPER_SGLINE_SYNTAX);
			return; 
		}
		
		if (sglines.count == 0) {
			notice_lang(s_OperServ, u, OPER_SGLINE_LIST_EMPTY);
			return;
		}
		
		if (isdigit(*mask) && strspn(mask, "1234567890,-") == strlen(mask)) {
			/* Deleting a range */
			res = slist_delete_range(&sglines, mask, NULL);
			if (res == 0) {
				notice_lang(s_OperServ, u, OPER_SGLINE_NO_MATCH);
				return;
			} else if (res == 1) {
				notice_lang(s_OperServ, u, OPER_SGLINE_DELETED_ONE);
			} else {
				notice_lang(s_OperServ, u, OPER_SGLINE_DELETED_SEVERAL, res);
			}
		} else {
			if ((res = slist_indexof(&sglines, mask)) == -1) {
				notice_lang(s_OperServ, u, OPER_SGLINE_NOT_FOUND, mask);
				return;	
			}
			
			slist_delete(&sglines, res);
			notice_lang(s_OperServ, u, OPER_SGLINE_DELETED, mask);
		}
		
		if (readonly)
			notice_lang(s_OperServ, u, READ_ONLY_MODE);
		
	} else if (!stricmp(cmd, "LIST")) {
		char *mask;
		int res, sent_header = 0;
	
		if (sglines.count == 0) {
			notice_lang(s_OperServ, u, OPER_SGLINE_LIST_EMPTY);
			return;
		}
		
		mask = strtok(NULL, "");
	
		if (!mask || (isdigit(*mask) && strspn(mask, "1234567890,-") == strlen(mask))) {
			res = slist_enum(&sglines, mask, &sgline_list_callback, u, &sent_header);
			if (res == 0) {
				notice_lang(s_OperServ, u, OPER_SGLINE_NO_MATCH);
				return;
			}
		} else {
			int i;
			char *amask;
			
			for (i = 0;	i < sglines.count; i++) {
				amask = ((SXLine *)sglines.list[i])->mask;
				if (!stricmp(mask, amask) || match_wild_nocase(mask, amask))
					sgline_list(i+1, sglines.list[i], u, &sent_header);
			}	
		
			if (!sent_header)
			notice_lang(s_OperServ, u, OPER_SGLINE_NO_MATCH);
		}
	} else if (!stricmp(cmd, "VIEW")) {
		char *mask;
		int res, sent_header = 0;
	
		if (sglines.count == 0) {
			notice_lang(s_OperServ, u, OPER_SGLINE_LIST_EMPTY);
			return;
		}
		
		mask = strtok(NULL, "");
	
		if (!mask || (isdigit(*mask) && strspn(mask, "1234567890,-") == strlen(mask))) {
			res = slist_enum(&sglines, mask, &sgline_view_callback, u, &sent_header);
			if (res == 0) {
				notice_lang(s_OperServ, u, OPER_SGLINE_NO_MATCH);
				return;
			}
		} else {
			int i;
			char *amask;
			
			for (i = 0;	i < sglines.count; i++) {
				amask = ((SXLine *)sglines.list[i])->mask;
				if (!stricmp(mask, amask) || match_wild_nocase(mask, amask))
					sgline_view(i+1, sglines.list[i], u, &sent_header);
			}	
		
			if (!sent_header)
			notice_lang(s_OperServ, u, OPER_SGLINE_NO_MATCH);
		}
	} else if (!stricmp(cmd, "CLEAR")) {
		slist_clear(&sglines, 1);
		notice_lang(s_OperServ, u, OPER_SGLINE_CLEAR);
	} else {
		syntax_error(s_OperServ, u, "SGLINE", OPER_SGLINE_SYNTAX);
	}
#else
	notice_lang(s_OperServ, u, OPER_SGLINE_UNSUPPORTED);
#endif
}

/*************************************************************************/

/* Adds an SQLINE to the list. Returns >= 0 on success, -1 if it failed, -2 if
 * only the expiry time changed.
 * The success result is the number of SQLINEs that were deleted to successfully add one.
 */

int add_sqline(User *u, char *mask, const char *by, const time_t expires, const char *reason)
{
	int deleted = 0, i;
	SXLine *entry;
	
	/* Checks whether there is an SQLINE that already covers
	 * the one we want to add, and whether there are SQLINEs
	 * that would be covered by this one.
	 * If so, warn the user in the first case and cleanup
	 * the useless SQLINEs in the second.
	 */
		
	if (sqlines.count > 0) {
			 
	for (i = sqlines.count - 1; i >= 0; i--) {
		entry = sqlines.list[i];
				
		if (!entry)
			continue;
				
		if (!stricmp(entry->mask, mask)) {
			if (entry->expires >= expires || entry->expires == 0) {
				if (u) notice_lang(s_OperServ, u, OPER_SQLINE_EXISTS, mask);
				return -1;
			} else {
				entry->expires = expires;
				if (u) notice_lang(s_OperServ, u, OPER_SQLINE_CHANGED, entry->mask);
				return -2;	
			}	
		}
				
		if (match_wild_nocase(entry->mask, mask) && (entry->expires >= expires || entry->expires == 0)) {
			if (u) notice_lang(s_OperServ, u, OPER_SQLINE_ALREADY_COVERED, mask, entry->mask);
			return -1;
		}
				
		if (match_wild_nocase(mask, entry->mask) && (entry->expires <= expires || expires == 0)) {
			slist_delete(&sqlines, i);
			deleted++;
		}
	}
	
	}
			
	/* We can now check whether the list is full or not. */
	if (slist_full(&sqlines)) {
		if (u) notice_lang(s_OperServ, u, OPER_SQLINE_REACHED_LIMIT, sqlines.limit);
		return -1;	
	}
			
	/* We can now (really) add the SQLINE. */
	entry = scalloc(sizeof(SXLine), 1);
	if (!entry) return -1;
	
	entry->mask = sstrdup(mask);
	entry->by = sstrdup(by);
	entry->reason = sstrdup(reason);
	entry->seton = time(NULL);
	entry->expires = expires;
			
	slist_add(&sqlines, entry);
	
	s_sqline(entry->mask, entry->reason);
	
	return deleted;
}

/* Does the user match any SQLINEs? */

int check_sqline(const char *nick)
{
	int i;
	SXLine *sx;
	
	if (sqlines.count == 0)
		return 0;
	
	for (i = 0; i < sqlines.count; i++) {
		sx = sqlines.list[i];
		if (!sx) 
			continue;
		
#ifdef IRC_BAHAMUT
		if (*sx->mask == '#')
			continue;
#endif
		
		if (match_wild_nocase(sx->mask, nick)) {
			s_sqline(sx->mask, sx->reason);
			/* We kill nick since s_sqline can't */
#ifdef IRC_BAHAMUT
			send_cmd(ServerName, "SVSKILL %s :Q-Lined: %s", nick, sx->reason);  
#else
			send_cmd(s_OperServ, "KILL %s :Q-Lined: %s", nick, sx->reason);
#endif
			return 1;
		}
	}
	
	return 0;
}

#ifdef IRC_BAHAMUT
int check_chan_sqline(const char *chan)
{
	int i;
	SXLine *sx;
	
	if (sqlines.count == 0)
		return 0;
	
	for (i = 0; i < sqlines.count; i++) {
		sx = sqlines.list[i];
		if (!sx) 
			continue;
		
		if (*sx->mask != '#')
			continue;
		
		if (match_wild_nocase(sx->mask, chan)) {
			s_sqline(sx->mask, sx->reason);
			return 1;
		}
	}
	
	return 0;
}
#endif

/* Delete any expired SQLINEs. */

void expire_sqlines(void)
{
    int i;
	time_t now = time(NULL);
	SXLine *sx;

	for (i = sqlines.count - 1; i >= 0; i--) {
		sx = sqlines.list[i];
		
		if (!sx->expires || sx->expires > now)
			continue;
			
		if (WallSQLineExpire)
	    	wallops(s_OperServ, "SQLINE on \2%s\2 has expired", sx->mask);
	    slist_delete(&sqlines, i);
	}
}

static void free_sqline_entry(SList *slist, void *item)
{
	SXLine *sx = item;
	
	/* Remove the SQLINE from all the servers */
	s_unsqline(sx->mask);
	
	/* Free the structure */
	free(sx->mask);
	free(sx->by);
	free(sx->reason);
	free(sx);
}

/* item1 is not an SXLine pointer, but a char */

static int is_sqline_entry_equal(SList *slist, void *item1, void *item2)
{
	char *sx1 = item1;
	SXLine *sx2 = item2;	
	
	if (!sx1 || !sx2) return 0;
	
	if (!stricmp(sx1, sx2->mask))
		return 1;
	else
		return 0;	
}

/* Lists an SQLINE entry, prefixing it with the header if needed */

static int sqline_list(int number, SXLine *sx, User *u, int *sent_header)
{
	if (!sx)
	return 0;
	
	if (!*sent_header) {
		notice_lang(s_OperServ, u, OPER_SQLINE_LIST_HEADER);
		*sent_header = 1;
	}
	
	notice_lang(s_OperServ, u, OPER_SQLINE_LIST_FORMAT, number, sx->mask, sx->reason);
		
	return 1; 	
}

/* Callback for enumeration purposes */

static int sqline_list_callback(SList *slist, int number, void *item, va_list args)
{
	User *u = va_arg(args, User *);
	int *sent_header = va_arg(args, int *);
	
	return sqline_list(number, item, u, sent_header);
}

/* Lists an SQLINE entry, prefixing it with the header if needed */

static int sqline_view(int number, SXLine *sx, User *u, int *sent_header)
{
	char timebuf[32], expirebuf[256];
	struct tm tm;
	
	if (!sx) 
	return 0;
	
	if (!*sent_header) {
		notice_lang(s_OperServ, u, OPER_SQLINE_VIEW_HEADER);
		*sent_header = 1;
	}
	
	tm = *localtime(&sx->seton);
	strftime_lang(timebuf, sizeof(timebuf), u, STRFTIME_SHORT_DATE_FORMAT, &tm);
	expire_left(u->na, expirebuf, sizeof(expirebuf), sx->expires);
	notice_lang(s_OperServ, u, OPER_SQLINE_VIEW_FORMAT, number, sx->mask, sx->by, timebuf, expirebuf, sx->reason);
	
	return 1; 	
}

/* Callback for enumeration purposes */

static int sqline_view_callback(SList *slist, int number, void *item, va_list args)
{
	User *u = va_arg(args, User *);
	int *sent_header = va_arg(args, int *);
	
	return sqline_view(number, item, u, sent_header);
}

/* Manage the SQLINE list. */

static void do_sqline(User *u)
{
	char *cmd = strtok(NULL, " ");
	
	if (!cmd)
	cmd = "";
	
	if (!stricmp(cmd, "ADD")) {
		int deleted = 0;
		char *expiry, *mask, *reason;
		time_t expires;
		
		mask = strtok(NULL, " ");
		if (mask && *mask == '+') {
			expiry = mask;
			mask = strtok(NULL, " ");
		} else {
			expiry = NULL;	
		}
		
		expires = expiry ? dotime(expiry) : SQLineExpiry;
		/* If the expiry given does not contain a final letter, it's in days,
		 * said the doc. Ah well.
		 */
		if (expiry && isdigit(expiry[strlen(expiry)-1]))
	    		expires *= 86400;
		/* Do not allow less than a minute expiry time */
		if (expires != 0 && expires < 60) {
	    	notice_lang(s_OperServ, u, BAD_EXPIRY_TIME);
	    	return;
		} else if (expires > 0) {
	    	expires += time(NULL);
		}
		
		if (mask && (reason = strtok(NULL, ""))) {
			
			/* We first do some sanity check on the proposed mask. */
			if (strspn(mask, "*?") == strlen(mask)) {
				notice_lang(s_OperServ, u, USERHOST_MASK_TOO_WIDE, mask);
				return;
			}
			
#ifndef IRC_BAHAMUT
			/* Channel SQLINEs are only supported on Bahamut servers */
			if (*mask == '#') {
				notice_lang(s_OperServ, u, OPER_SQLINE_CHANNELS_UNSUPPORTED);
				return;
			}
#endif
			
			deleted = add_sqline(u, mask, u->nick, expires, reason);
			if (deleted < 0)
				return;
			else if (deleted)
				notice_lang(s_OperServ, u, OPER_SQLINE_DELETED_SEVERAL, deleted);
			notice_lang(s_OperServ, u, OPER_SQLINE_ADDED, mask); 
			
			if (WallOSSQLine) {
				char buf[128];
				
				if(!expires) {
					strcpy(buf, "does not expire");
				} else {
					int wall_expiry = expires-time(NULL);
					char *s = NULL;
					
					if (wall_expiry >= 86400) {
						wall_expiry /= 86400;
						s = "day";
					} else if (wall_expiry >= 3600) { 
						wall_expiry /= 3600;
						s = "hour";
					} else if (wall_expiry >= 60) {
						wall_expiry /= 60;
						s = "minute";
					}
					
					snprintf(buf, sizeof(buf), "expires in %d %s%s", wall_expiry, s, (wall_expiry == 1) ? "" : "s");
				}
				
				wallops(s_OperServ, "%s added an SQLINE for %s (%s)", u->nick, mask, buf);
			}
			
			if (readonly)
				notice_lang(s_OperServ, u, READ_ONLY_MODE);
			
		} else {
			syntax_error(s_OperServ, u, "SQLINE", OPER_SQLINE_SYNTAX);
		}
		
	} else if (!stricmp(cmd, "DEL")) {
		
		char *mask;
		int res = 0;
		
		mask = strtok(NULL, "");
		
		if (!mask) {
			syntax_error(s_OperServ, u, "SQLINE", OPER_SQLINE_SYNTAX);
			return; 
		}
		
		if (sqlines.count == 0) {
			notice_lang(s_OperServ, u, OPER_SQLINE_LIST_EMPTY);
			return;
		}
		
		if (isdigit(*mask) && strspn(mask, "1234567890,-") == strlen(mask)) {
			/* Deleting a range */
			res = slist_delete_range(&sqlines, mask, NULL);
			if (res == 0) {
				notice_lang(s_OperServ, u, OPER_SQLINE_NO_MATCH);
				return;
			} else if (res == 1) {
				notice_lang(s_OperServ, u, OPER_SQLINE_DELETED_ONE);
			} else {
				notice_lang(s_OperServ, u, OPER_SQLINE_DELETED_SEVERAL, res);
			}
		} else {
			if ((res = slist_indexof(&sqlines, mask)) == -1) {
				notice_lang(s_OperServ, u, OPER_SQLINE_NOT_FOUND, mask);
				return;	
			}
			
			slist_delete(&sqlines, res);
			notice_lang(s_OperServ, u, OPER_SQLINE_DELETED, mask);
		}
		
		if (readonly)
			notice_lang(s_OperServ, u, READ_ONLY_MODE);
		
	} else if (!stricmp(cmd, "LIST")) {
		char *mask;
		int res, sent_header = 0;
	
		if (sqlines.count == 0) {
			notice_lang(s_OperServ, u, OPER_SQLINE_LIST_EMPTY);
			return;
		}
		
		mask = strtok(NULL, "");
	
		if (!mask || (isdigit(*mask) && strspn(mask, "1234567890,-") == strlen(mask))) {
			res = slist_enum(&sqlines, mask, &sqline_list_callback, u, &sent_header);
			if (res == 0) {
				notice_lang(s_OperServ, u, OPER_SQLINE_NO_MATCH);
				return;
			}
		} else {
			int i;
			char *amask;
			
			for (i = 0;	i < sqlines.count; i++) {
				amask = ((SXLine *)sqlines.list[i])->mask;
				if (!stricmp(mask, amask) || match_wild_nocase(mask, amask))
					sqline_list(i+1, sqlines.list[i], u, &sent_header);
			}	
		
			if (!sent_header)
			notice_lang(s_OperServ, u, OPER_SQLINE_NO_MATCH);
		}
	} else if (!stricmp(cmd, "VIEW")) {
		char *mask;
		int res, sent_header = 0;
	
		if (sqlines.count == 0) {
			notice_lang(s_OperServ, u, OPER_SQLINE_LIST_EMPTY);
			return;
		}
		
		mask = strtok(NULL, "");
	
		if (!mask || (isdigit(*mask) && strspn(mask, "1234567890,-") == strlen(mask))) {
			res = slist_enum(&sqlines, mask, &sqline_view_callback, u, &sent_header);
			if (res == 0) {
				notice_lang(s_OperServ, u, OPER_SQLINE_NO_MATCH);
				return;
			}
		} else {
			int i;
			char *amask;
			
			for (i = 0;	i < sqlines.count; i++) {
				amask = ((SXLine *)sqlines.list[i])->mask;
				if (!stricmp(mask, amask) || match_wild_nocase(mask, amask))
					sqline_view(i+1, sqlines.list[i], u, &sent_header);
			}	
		
			if (!sent_header)
			notice_lang(s_OperServ, u, OPER_SQLINE_NO_MATCH);
		}
	} else if (!stricmp(cmd, "CLEAR")) {
		slist_clear(&sqlines, 1);
		notice_lang(s_OperServ, u, OPER_SQLINE_CLEAR);
	} else {
		syntax_error(s_OperServ, u, "SQLINE", OPER_SQLINE_SYNTAX);
	}
}

/*************************************************************************/

#ifdef IRC_BAHAMUT

/* Adds an SZLINE to the list. Returns >= 0 on success, -1 on error, -2 if
 * only the expiry time changed.
 * The success result is the number of SZLINEs that were deleted to successfully add one.
 */

int add_szline(User *u, char *mask, const char *by, const time_t expires, const char *reason)
{
	int deleted = 0, i;
	SXLine *entry;
			
	/* Checks whether there is an SZLINE that already covers
	 * the one we want to add, and whether there are SZLINEs
	 * that would be covered by this one.
	 * If so, warn the user in the first case and cleanup
	 * the useless SZLINEs in the second.
	 */
		
	if (szlines.count > 0) {
			 
	for (i = szlines.count - 1; i >= 0; i--) {
		entry = szlines.list[i];
				
		if (!entry)
			continue;
				
		if (!stricmp(entry->mask, mask)) {
			if (entry->expires >= expires || entry->expires == 0) {
				if (u) notice_lang(s_OperServ, u, OPER_SZLINE_EXISTS, mask);
				return -1;
			} else {
				entry->expires = expires;
				if (u) notice_lang(s_OperServ, u, OPER_SZLINE_EXISTS, mask);
				return -2;
			}
		}
				
		if (match_wild_nocase(entry->mask, mask)) {
			if (u) notice_lang(s_OperServ, u, OPER_SZLINE_ALREADY_COVERED, mask, entry->mask);
			return -1;	
		}
				
		if (match_wild_nocase(mask, entry->mask)) {
			slist_delete(&szlines, i);
			deleted++;
		}
	}
	
	}
			
	/* We can now check whether the list is full or not. */
	if (slist_full(&szlines)) {
		if (u) notice_lang(s_OperServ, u, OPER_SZLINE_REACHED_LIMIT, szlines.limit);
		return -1;
	}
			
	/* We can now (really) add the SZLINE. */
	entry = scalloc(sizeof(SXLine), 1);
	if (!entry) return -1;
	
	entry->mask = sstrdup(mask);
	entry->by = sstrdup(by);
	entry->reason = sstrdup(reason);
	entry->seton = time(NULL);
	entry->expires = expires;
			
	slist_add(&szlines, entry);
	s_szline(entry->mask, entry->reason);
	
	return deleted;
}

/* Delete any expired SZLINEs. */

void expire_szlines(void)
{
    int i;
	time_t now = time(NULL);
	SXLine *sx;

	for (i = szlines.count - 1; i >= 0; i--) {
		sx = szlines.list[i];
		
		if (!sx->expires || sx->expires > now)
			continue;
			
		if (WallSZLineExpire)
	    	wallops(s_OperServ, "SZLINE on \2%s\2 has expired", sx->mask);
	    slist_delete(&szlines, i);
	}
}

static void free_szline_entry(SList *slist, void *item)
{
	SXLine *sx = item;
	
	/* Remove the SZLINE from all the servers */
	s_unszline(sx->mask);
	
	/* Free the structure */
	free(sx->mask);
	free(sx->by);
	free(sx->reason);
	free(sx);
}

/* item1 is not an SXLine pointer, but a char 
 */

static int is_szline_entry_equal(SList *slist, void *item1, void *item2)
{
	char *sx1 = item1;
	SXLine *sx2 = item2;	
	
	if (!sx1 || !sx2) return 0;
	
	if (!stricmp(sx1, sx2->mask))
		return 1;
	else
		return 0;	
}

/* Lists an SZLINE entry, prefixing it with the header if needed */

static int szline_list(int number, SXLine *sx, User *u, int *sent_header)
{
	if (!sx) 
	return 0;
	
	if (!*sent_header) {
		notice_lang(s_OperServ, u, OPER_SZLINE_LIST_HEADER);
		*sent_header = 1;
	}
	
	notice_lang(s_OperServ, u, OPER_SZLINE_LIST_FORMAT, number, sx->mask, sx->reason);
		
	return 1; 	
}

/* Callback for enumeration purposes */

static int szline_list_callback(SList *slist, int number, void *item, va_list args)
{
	User *u = va_arg(args, User *);
	int *sent_header = va_arg(args, int *);
	
	return szline_list(number, item, u, sent_header);
}

/* Lists an SZLINE entry, prefixing it with the header if needed */

static int szline_view(int number, SXLine *sx, User *u, int *sent_header)
{
	char timebuf[32], expirebuf[256];
	struct tm tm;
	
	if (!sx) 
	return 0;
	
	if (!*sent_header) {
		notice_lang(s_OperServ, u, OPER_SZLINE_VIEW_HEADER);
		*sent_header = 1;
	}
	
	tm = *localtime(&sx->seton);
	strftime_lang(timebuf, sizeof(timebuf), u, STRFTIME_SHORT_DATE_FORMAT, &tm);
	expire_left(u->na, expirebuf, sizeof(expirebuf), sx->expires);
	notice_lang(s_OperServ, u, OPER_SZLINE_VIEW_FORMAT, number, sx->mask, sx->by, timebuf, expirebuf, sx->reason);
		
	return 1; 	
}

/* Callback for enumeration purposes */

static int szline_view_callback(SList *slist, int number, void *item, va_list args)
{
	User *u = va_arg(args, User *);
	int *sent_header = va_arg(args, int *);
	
	return szline_view(number, item, u, sent_header);
}

#endif

/* Manage the SZLINE list. */

static void do_szline(User *u)
{
#ifdef IRC_BAHAMUT
	char *cmd = strtok(NULL, " ");
	
	if (!cmd)
	cmd = "";
	
	if (!stricmp(cmd, "ADD")) {
		int deleted = 0;
		char *expiry, *mask, *reason;
		time_t expires;
		
		mask = strtok(NULL, " ");
		if (mask && *mask == '+') {
			expiry = mask;
			mask = strtok(NULL, " ");
		} else {
			expiry = NULL;	
		}
		
		expires = expiry ? dotime(expiry) : SZLineExpiry;
		/* If the expiry given does not contain a final letter, it's in days,
		 * said the doc. Ah well.
		 */
		if (expiry && isdigit(expiry[strlen(expiry)-1]))
	    		expires *= 86400;
		/* Do not allow less than a minute expiry time */
		if (expires != 0 && expires < 60) {
	    	notice_lang(s_OperServ, u, BAD_EXPIRY_TIME);
	    	return;
		} else if (expires > 0) {
	    	expires += time(NULL);
		}
		
		if (mask && (reason = strtok(NULL, ""))) {
			/* We first do some sanity check on the proposed mask. */
			
			if (strchr(mask, '!') || strchr(mask, '@')) {
				notice_lang(s_OperServ, u, OPER_SZLINE_ONLY_IPS);
				return;
			}
			
			if (strspn(mask, "*?") == strlen(mask)) {
				notice_lang(s_OperServ, u, USERHOST_MASK_TOO_WIDE, mask);
				return;
			}
			
			deleted = add_szline(u, mask, u->nick, expires, reason);
			if (deleted < 0)
				return;
			else if (deleted)
				notice_lang(s_OperServ, u, OPER_SZLINE_DELETED_SEVERAL, deleted);
			notice_lang(s_OperServ, u, OPER_SZLINE_ADDED, mask); 
			
			if (WallOSSZLine) {
				char buf[128];
				
				if(!expires) {
					strcpy(buf, "does not expire");
				} else {
					int wall_expiry = expires-time(NULL);
					char *s = NULL;
					
					if (wall_expiry >= 86400) {
						wall_expiry /= 86400;
						s = "day";
					} else if (wall_expiry >= 3600) { 
						wall_expiry /= 3600;
						s = "hour";
					} else if (wall_expiry >= 60) {
						wall_expiry /= 60;
						s = "minute";
					}
					
					snprintf(buf, sizeof(buf), "expires in %d %s%s", wall_expiry, s, (wall_expiry == 1) ? "" : "s");
				}
				
				wallops(s_OperServ, "%s added an SZLINE for %s (%s)", u->nick, mask, buf);
			}
			
			if (readonly)
				notice_lang(s_OperServ, u, READ_ONLY_MODE);
			
		} else {
			syntax_error(s_OperServ, u, "SZLINE", OPER_SZLINE_SYNTAX);
		}
		
	} else if (!stricmp(cmd, "DEL")) {
		
		char *mask;
		int res = 0;
		
		mask = strtok(NULL, " ");
		
		if (!mask) {
			syntax_error(s_OperServ, u, "SZLINE", OPER_SZLINE_SYNTAX);
			return; 
		}
		
		if (szlines.count == 0) {
			notice_lang(s_OperServ, u, OPER_SZLINE_LIST_EMPTY);
			return;
		}
		
		if (isdigit(*mask) && strspn(mask, "1234567890,-") == strlen(mask)) {
			/* Deleting a range */
			res = slist_delete_range(&szlines, mask, NULL);
			if (res == 0) {
				notice_lang(s_OperServ, u, OPER_SZLINE_NO_MATCH);
				return;
			} else if (res == 1) {
				notice_lang(s_OperServ, u, OPER_SZLINE_DELETED_ONE);
			} else {
				notice_lang(s_OperServ, u, OPER_SZLINE_DELETED_SEVERAL, res);
			}
		} else {
			if ((res = slist_indexof(&szlines, mask)) == -1) {
				notice_lang(s_OperServ, u, OPER_SZLINE_NOT_FOUND, mask);
				return;	
			}
			
			slist_delete(&szlines, res);
			notice_lang(s_OperServ, u, OPER_SZLINE_DELETED, mask);
		}
		
		if (readonly)
			notice_lang(s_OperServ, u, READ_ONLY_MODE);
		
	} else if (!stricmp(cmd, "LIST")) {
		char *mask;
		int res, sent_header = 0;
	
		if (szlines.count == 0) {
			notice_lang(s_OperServ, u, OPER_SZLINE_LIST_EMPTY);
			return;
		}
		
		mask = strtok(NULL, " ");
	
		if (!mask || (isdigit(*mask) && strspn(mask, "1234567890,-") == strlen(mask))) {
			res = slist_enum(&szlines, mask, &szline_list_callback, u, &sent_header);
			if (res == 0) {
				notice_lang(s_OperServ, u, OPER_SZLINE_NO_MATCH);
				return;
			}
		} else {
			int i;
			char *amask;
			
			for (i = 0;	i < szlines.count; i++) {
				amask = ((SXLine *)szlines.list[i])->mask;
				if (!stricmp(mask, amask) || match_wild_nocase(mask, amask))
					szline_list(i+1, szlines.list[i], u, &sent_header);
			}	
		
			if (!sent_header)
			notice_lang(s_OperServ, u, OPER_SZLINE_NO_MATCH);
		}
	} else if (!stricmp(cmd, "VIEW")) {
		char *mask;
		int res, sent_header = 0;
	
		if (szlines.count == 0) {
			notice_lang(s_OperServ, u, OPER_SZLINE_LIST_EMPTY);
			return;
		}
		
		mask = strtok(NULL, " ");
	
		if (!mask || (isdigit(*mask) && strspn(mask, "1234567890,-") == strlen(mask))) {
			res = slist_enum(&szlines, mask, &szline_view_callback, u, &sent_header);
			if (res == 0) {
				notice_lang(s_OperServ, u, OPER_SZLINE_NO_MATCH);
				return;
			}
		} else {
			int i;
			char *amask;
			
			for (i = 0;	i < szlines.count; i++) {
				amask = ((SXLine *)szlines.list[i])->mask;
				if (!stricmp(mask, amask) || match_wild_nocase(mask, amask))
					szline_view(i+1, szlines.list[i], u, &sent_header);
			}	
		
			if (!sent_header)
			notice_lang(s_OperServ, u, OPER_SZLINE_NO_MATCH);
		}
	} else if (!stricmp(cmd, "CLEAR")) {
		slist_clear(&szlines, 1);
		notice_lang(s_OperServ, u, OPER_SZLINE_CLEAR);
	} else {
		syntax_error(s_OperServ, u, "SZLINE", OPER_SZLINE_SYNTAX);
	}
#else
	notice_lang(s_OperServ, u, OPER_SZLINE_UNSUPPORTED);
#endif
}

/*************************************************************************/

static void do_chanlist(User *u)
{
	char *pattern = strtok(NULL, " ");
	char *opt = strtok(NULL, " ");
	
	int modes = 0;
	User *u2;
		
	if (opt && !stricmp(opt, "SECRET"))
		modes |= (CMODE_s | CMODE_p);
		
	if (pattern && (u2 = finduser(pattern))) {
		struct u_chanlist *uc;
		
		notice_lang(s_OperServ, u, OPER_CHANLIST_HEADER_USER, u2->nick);
		
		for (uc = u2->chans; uc; uc = uc->next) {
			if (modes && !(uc->chan->mode & modes)) 
				continue;
			notice_lang(s_OperServ, u, OPER_CHANLIST_RECORD, uc->chan->name, uc->chan->usercount, chan_get_modes(uc->chan, 1), (uc->chan->topic ? uc->chan->topic : ""));	
		}
	} else {
		int i;
		Channel *c;
	
		notice_lang(s_OperServ, u, OPER_CHANLIST_HEADER);
	
		for (i = 0; i < 1024; i++) {
			for (c = chanlist[i]; c; c = c->next) {
				if (pattern && !match_wild_nocase(pattern, c->name))
					continue;
				if (modes && !(c->mode & modes)) 
					continue;
				notice_lang(s_OperServ, u, OPER_CHANLIST_RECORD, c->name, c->usercount, chan_get_modes(c, 1), (c->topic ? c->topic : ""));
			}
		}
	}
	
	notice_lang(s_OperServ, u, OPER_CHANLIST_END);
}

/*************************************************************************/

static void do_userlist(User *u)
{
	char *pattern = strtok(NULL, " ");
	char *opt = strtok(NULL, " ");
	
	Channel *c;
	int modes = 0;
		
	if (opt && !stricmp(opt, "INVISIBLE"))
		modes |= UMODE_i;
	
	if (pattern && (c = findchan(pattern))) {
		struct c_userlist *cu;
		
		notice_lang(s_OperServ, u, OPER_USERLIST_HEADER_CHAN, pattern);
		
		for (cu = c->users; cu; cu = cu->next) {
			if (modes && !(cu->user->mode & modes)) 
				continue;
			notice_lang(s_OperServ, u, OPER_USERLIST_RECORD, cu->user->nick, cu->user->username, GetHost(cu->user));
		}
	} else {
		char mask[BUFSIZE];
		int i;
		User *u2;
		
		notice_lang(s_OperServ, u, OPER_USERLIST_HEADER);
	
		for (i = 0; i < 1024; i++) {
			for (u2 = userlist[i]; u2; u2 = u2->next) {
				if (pattern) {
					snprintf(mask, sizeof(mask), "%s!%s@%s", u2->nick, u2->username, GetHost(u2));
					if (!match_wild_nocase(pattern, mask)) continue;
					if (modes && !(u2->mode & modes)) continue;
				}
				notice_lang(s_OperServ, u, OPER_USERLIST_RECORD, u2->nick, u2->username, GetHost(u2));
			}
		}
	}
	
	notice_lang(s_OperServ, u, OPER_USERLIST_END);
}

/*************************************************************************/

/* Callback function used to sort the admin list */

static int compare_adminlist_entries(SList *slist, void *item1, void *item2)
{
	NickCore *nc1 = item1, *nc2 = item2;
	if (!nc1 || !nc2) return -1;		/* To tell to continue */
	return stricmp(nc1->display, nc2->display);	
}

/* Callback function used when an admin list entry is deleted */

static void free_adminlist_entry(SList *slist, void *item)
{
	NickCore *nc = item;
	nc->flags &= ~NI_SERVICES_ADMIN;
}

/* Lists an admin entry, prefixing it with the header if needed */

static int admin_list(int number, NickCore *nc, User *u, int *sent_header)
{
	if (!nc) 
	return 0;
	
	if (!*sent_header) {
		notice_lang(s_OperServ, u, OPER_ADMIN_LIST_HEADER);
		*sent_header = 1;
	}
	
	notice_lang(s_OperServ, u, OPER_ADMIN_LIST_FORMAT, number, nc->display);
	return 1; 	
}

/* Callback for enumeration purposes */

static int admin_list_callback(SList *slist, int number, void *item, va_list args)
{
	User *u = va_arg(args, User *);
	int *sent_header = va_arg(args, int *);
	
	return admin_list(number, item, u, sent_header);
}

/* Services admin list viewing/modification. */

static void do_admin(User *u)
{
	char *cmd = strtok(NULL, " ");
	char *nick = strtok(NULL, " ");
	NickAlias *na;
	int res = 0;
	
	if (skeleton) {
	notice_lang(s_OperServ, u, OPER_ADMIN_SKELETON);
	return;
    }
	
	if (!cmd || (!nick && stricmp(cmd, "LIST") && stricmp(cmd, "CLEAR"))) {
		syntax_error(s_OperServ, u, "ADMIN", OPER_ADMIN_SYNTAX);
	} else if (!stricmp(cmd, "ADD")) {
		if (!is_services_root(u)) {
			notice_lang(s_OperServ, u, PERMISSION_DENIED);
			return;
		}

		if (!(na = findnick(nick))) {
			notice_lang(s_OperServ, u, NICK_X_NOT_REGISTERED, nick);
			return;
		}
		
		if (na->status & NS_VERBOTEN) {
			notice_lang(s_OperServ, u, NICK_X_FORBIDDEN, nick);
			return;	
		}
		
		if (na->nc->flags & NI_SERVICES_ADMIN || slist_indexof(&servadmins, na->nc) != -1) {
			notice_lang(s_OperServ, u, OPER_ADMIN_EXISTS, nick);
			return;
		}
		
		res = slist_add(&servadmins, na->nc);
		if (res == -2) {
			notice_lang(s_OperServ, u, OPER_ADMIN_REACHED_LIMIT, nick);
			return;
		} else  {
			na->nc->flags |= NI_SERVICES_ADMIN;
			notice_lang(s_OperServ, u, OPER_ADMIN_ADDED, nick);
		}
		
		if (readonly)
		notice_lang(s_OperServ, u, READ_ONLY_MODE);
	} else if (!stricmp(cmd, "DEL")) {
		if (!is_services_root(u)) {
			notice_lang(s_OperServ, u, PERMISSION_DENIED);
			return;
		}
		
		if (servadmins.count == 0) {
			notice_lang(s_OperServ, u, OPER_ADMIN_LIST_EMPTY);
			return;
		}
		
		if (isdigit(*nick) && strspn(nick, "1234567890,-") == strlen(nick)) {
			/* Deleting a range */
			res = slist_delete_range(&servadmins, nick, NULL);
			if (res == 0) {
				notice_lang(s_OperServ, u, OPER_ADMIN_NO_MATCH);
				return;
			} else if (res == 1) {
				notice_lang(s_OperServ, u, OPER_ADMIN_DELETED_ONE);
			} else {
				notice_lang(s_OperServ, u, OPER_ADMIN_DELETED_SEVERAL, res);
			}
		} else {
			if (!(na = findnick(nick))) {
				notice_lang(s_OperServ, u, NICK_X_NOT_REGISTERED, nick);
				return;
			}
		
			if (na->status & NS_VERBOTEN) {
				notice_lang(s_OperServ, u, NICK_X_FORBIDDEN, nick);
				return;	
			}
			
			if (!(na->nc->flags & NI_SERVICES_ADMIN) || (res = slist_indexof(&servadmins, na->nc)) == -1) {
				notice_lang(s_OperServ, u, OPER_ADMIN_NOT_FOUND, nick);
				return;	
			}
			
			slist_delete(&servadmins, res);
			notice_lang(s_OperServ, u, OPER_ADMIN_DELETED, nick);
		}
		
		if (readonly)
		notice_lang(s_OperServ, u, READ_ONLY_MODE);
	} else if (!stricmp(cmd, "LIST")) {
		int sent_header = 0;
	
		if (servadmins.count == 0) {
			notice_lang(s_OperServ, u, OPER_ADMIN_LIST_EMPTY);
			return;
		}
	
		if (!nick || (isdigit(*nick) && strspn(nick, "1234567890,-") == strlen(nick))) {
			res = slist_enum(&servadmins, nick, &admin_list_callback, u, &sent_header);
			if (res == 0) {
				notice_lang(s_OperServ, u, OPER_ADMIN_NO_MATCH);
				return;
			}
		} else {
			int i;
			
			for (i = 0;	i < servadmins.count; i++)
				if (!stricmp(nick, ((NickCore *)servadmins.list[i])->display) || match_wild_nocase(nick, ((NickCore *)servadmins.list[i])->display))
					admin_list(i+1, servadmins.list[i], u, &sent_header);
			
			if (!sent_header)
			notice_lang(s_OperServ, u, OPER_ADMIN_NO_MATCH);
		}
	} else if (!stricmp(cmd, "CLEAR")) {
		if (!is_services_root(u)) {
			notice_lang(s_OperServ, u, PERMISSION_DENIED);
			return;
		}
		
		if (servadmins.count == 0) {
			notice_lang(s_OperServ, u, OPER_ADMIN_LIST_EMPTY);
			return;
		}
		
		slist_clear(&servadmins, 1);
		notice_lang(s_OperServ, u, OPER_ADMIN_CLEAR);
	} else {
		syntax_error(s_OperServ, u, "ADMIN", OPER_ADMIN_SYNTAX);	
	}
}

/*************************************************************************/

/* Callback function used to sort the oper list */

static int compare_operlist_entries(SList *slist, void *item1, void *item2)
{
	NickCore *nc1 = item1, *nc2 = item2;
	if (!nc1 || !nc2) return -1;		/* To tell to continue */
	return stricmp(nc1->display, nc2->display);
}

/* Callback function used when an oper list entry is deleted */

static void free_operlist_entry(SList *slist, void *item)
{
	NickCore *nc = item;
	nc->flags &= ~NI_SERVICES_OPER;
}

/* Lists an oper entry, prefixing it with the header if needed */

static int oper_list(int number, NickCore *nc, User *u, int *sent_header)
{
	if (!nc) 
	return 0;
	
	if (!*sent_header) {
		notice_lang(s_OperServ, u, OPER_OPER_LIST_HEADER);
		*sent_header = 1;
	}
	
	notice_lang(s_OperServ, u, OPER_OPER_LIST_FORMAT, number, nc->display);
	return 1; 	
}

/* Callback for enumeration purposes */

static int oper_list_callback(SList *slist, int number, void *item, va_list args)
{
	User *u = va_arg(args, User *);
	int *sent_header = va_arg(args, int *);
	
	return oper_list(number, item, u, sent_header);
}

/* Services operator list viewing/modification. */

static void do_oper(User *u)
{
	char *cmd = strtok(NULL, " ");
	char *nick = strtok(NULL, " ");
	NickAlias *na;
	int res = 0;
	
	if (skeleton) {
	notice_lang(s_OperServ, u, OPER_OPER_SKELETON);
	return;
    }
	
	if (!cmd || (!nick && stricmp(cmd, "LIST") && stricmp(cmd, "CLEAR"))) {
		syntax_error(s_OperServ, u, "OPER", OPER_OPER_SYNTAX);
	} else if (!stricmp(cmd, "ADD")) {
		if (!is_services_admin(u)) {
			notice_lang(s_OperServ, u, PERMISSION_DENIED);
			return;
		}
	
		if (!(na = findnick(nick))) {
			notice_lang(s_OperServ, u, NICK_X_NOT_REGISTERED, nick);
			return;
		}
		
		if (na->status & NS_VERBOTEN) {
			notice_lang(s_OperServ, u, NICK_X_FORBIDDEN, nick);
			return;	
		}
		
		if (na->nc->flags & NI_SERVICES_OPER || slist_indexof(&servopers, na->nc) != -1) {
			notice_lang(s_OperServ, u, OPER_OPER_EXISTS, nick);
			return;	
		}
		
		res = slist_add(&servopers, na->nc);
		if (res == -2) {
			notice_lang(s_OperServ, u, OPER_OPER_REACHED_LIMIT, nick);
			return;
		} else  {
			na->nc->flags |= NI_SERVICES_OPER;
			notice_lang(s_OperServ, u, OPER_OPER_ADDED, nick);
		}
		
		if (readonly)
		notice_lang(s_OperServ, u, READ_ONLY_MODE);
	} else if (!stricmp(cmd, "DEL")) {
		if (!is_services_admin(u)) {
			notice_lang(s_OperServ, u, PERMISSION_DENIED);
			return;
		}
		
		if (isdigit(*nick) && strspn(nick, "1234567890,-") == strlen(nick)) {
			/* Deleting a range */
			res = slist_delete_range(&servopers, nick, NULL);
			if (res == 0) {
				notice_lang(s_OperServ, u, OPER_OPER_NO_MATCH);
				return;
			} else if (res == 1) {
				notice_lang(s_OperServ, u, OPER_OPER_DELETED_ONE);
			} else {
				notice_lang(s_OperServ, u, OPER_OPER_DELETED_SEVERAL, res);
			}
		} else {
			if (!(na = findnick(nick))) {
				notice_lang(s_OperServ, u, NICK_X_NOT_REGISTERED, nick);
				return;
			}
		
			if (na->status & NS_VERBOTEN) {
				notice_lang(s_OperServ, u, NICK_X_FORBIDDEN, nick);
				return;	
			}
			
			if (!(na->nc->flags & NI_SERVICES_OPER) || (res = slist_indexof(&servopers, na->nc)) == -1) {
				notice_lang(s_OperServ, u, OPER_OPER_NOT_FOUND, nick);
				return;	
			}
			
			slist_delete(&servopers, res);
			notice_lang(s_OperServ, u, OPER_OPER_DELETED, nick);
		}
		
		if (readonly)
		notice_lang(s_OperServ, u, READ_ONLY_MODE);
	} else if (!stricmp(cmd, "LIST")) {
		int sent_header = 0;
	
		if (servopers.count == 0) {
			notice_lang(s_OperServ, u, OPER_OPER_LIST_EMPTY);
			return;
		}
	
		if (!nick || (isdigit(*nick) && strspn(nick, "1234567890,-") == strlen(nick))) {
			res = slist_enum(&servopers, nick, &oper_list_callback, u, &sent_header);
			if (res == 0) {
				notice_lang(s_OperServ, u, OPER_OPER_NO_MATCH);
				return;
			}
		} else {
			int i;
			
			for (i = 0;	i < servopers.count; i++)
				if (!stricmp(nick, ((NickCore *)servopers.list[i])->display) || match_wild_nocase(nick, ((NickCore *)servopers.list[i])->display))
					oper_list(i+1, servopers.list[i], u, &sent_header);
			
			if (!sent_header)
			notice_lang(s_OperServ, u, OPER_OPER_NO_MATCH);
		}
	} else if (!stricmp(cmd, "CLEAR")) {
		if (!is_services_admin(u)) {
			notice_lang(s_OperServ, u, PERMISSION_DENIED);
			return;
		}
		
		if (servopers.count == 0) {
			notice_lang(s_OperServ, u, OPER_OPER_LIST_EMPTY);
			return;
		}
		
		slist_clear(&servopers, 1);
		notice_lang(s_OperServ, u, OPER_OPER_CLEAR);
	} else {
		syntax_error(s_OperServ, u, "OPER", OPER_OPER_SYNTAX);	
	}
}

/*************************************************************************/

/* Set various Services runtime options. */

static void do_set(User *u)
{
    char *option = strtok(NULL, " ");
    char *setting = strtok(NULL, " ");

    if (!option || !setting) {
	syntax_error(s_OperServ, u, "SET", OPER_SET_SYNTAX);

    } else if (stricmp(option, "IGNORE") == 0) {
	if (stricmp(setting, "on") == 0) {
	    allow_ignore = 1;
	    notice_lang(s_OperServ, u, OPER_SET_IGNORE_ON);
	} else if (stricmp(setting, "off") == 0) {
	    allow_ignore = 0;
	    notice_lang(s_OperServ, u, OPER_SET_IGNORE_OFF);
	} else {
	    notice_lang(s_OperServ, u, OPER_SET_IGNORE_ERROR);
	}

    } else if (stricmp(option, "READONLY") == 0) {
	if (stricmp(setting, "on") == 0) {
	    readonly = 1;
	    log("Read-only mode activated");
	    close_log();
	    notice_lang(s_OperServ, u, OPER_SET_READONLY_ON);
	} else if (stricmp(setting, "off") == 0) {
	    readonly = 0;
	    open_log();
	    log("Read-only mode deactivated");
	    notice_lang(s_OperServ, u, OPER_SET_READONLY_OFF);
	} else {
	    notice_lang(s_OperServ, u, OPER_SET_READONLY_ERROR);
	}

    } else if (stricmp(option, "DEBUG") == 0) {
	if (stricmp(setting, "on") == 0) {
	    debug = 1;
	    log("Debug mode activated");
	    notice_lang(s_OperServ, u, OPER_SET_DEBUG_ON);
	} else if (stricmp(setting, "off") == 0 ||
				(*setting == '0' && atoi(setting) == 0)) {
	    log("Debug mode deactivated");
	    debug = 0;
	    notice_lang(s_OperServ, u, OPER_SET_DEBUG_OFF);
	} else if (isdigit(*setting) && atoi(setting) > 0) {
	    debug = atoi(setting);
	    log("Debug mode activated (level %d)", debug);
	    notice_lang(s_OperServ, u, OPER_SET_DEBUG_LEVEL, debug);
	} else {
	    notice_lang(s_OperServ, u, OPER_SET_DEBUG_ERROR);
	}
	
	} else if (stricmp(option, "NOEXPIRE") == 0) {
	if (stricmp(setting, "ON") == 0) {
	    noexpire = 1;
	    log("No expire mode activated");
	    notice_lang(s_OperServ, u, OPER_SET_NOEXPIRE_ON);
	} else if (stricmp(setting, "OFF") == 0) {
	    noexpire = 0;
	    log("No expire mode deactivated");
	    notice_lang(s_OperServ, u, OPER_SET_NOEXPIRE_OFF);
	} else {
	    notice_lang(s_OperServ, u, OPER_SET_NOEXPIRE_ERROR);
	}

    } else {
	notice_lang(s_OperServ, u, OPER_SET_UNKNOWN_OPTION, option);
    }
}

/*************************************************************************/

static void do_noop(User *u)
{
	char *cmd = strtok(NULL, " ");
    char *server = strtok(NULL, " ");

	if (!cmd || !server) {
		syntax_error(s_OperServ, u, "NOOP", OPER_NOOP_SYNTAX);
	} else if (!stricmp(cmd, "SET")) {
		User *u2;
		char reason[NICKMAX+32];
		
		/* Remove the O:lines */
		s_svsnoop(server, 1);
		
		/* Kill all the IRCops of the server */
		snprintf(reason, sizeof(reason), "NOOP command used by %s", u->nick);
		for (u2 = firstuser(); u2; u2 = nextuser())
			if (is_oper(u2) && !stricmp(u2->server, server))
#ifdef IRC_BAHAMUT
				send_cmd(ServerName, "SVSKILL %s :%s", u2->nick, reason);
#else
				kill_user(s_OperServ, u2->nick, reason);
#endif

		if (WallOSNoOp)
			wallops(s_OperServ, "\2%s\2 used NOOP on \2%s\2", u->nick, server);
		notice_lang(s_OperServ, u, OPER_NOOP_SET, server);
	} else if (!stricmp(cmd, "REVOKE")) {
		s_svsnoop(server, 0);
		notice_lang(s_OperServ, u, OPER_NOOP_REVOKE, server);
	} else {
		syntax_error(s_OperServ, u, "NOOP", OPER_NOOP_SYNTAX);
	}
}

/*************************************************************************/

static void do_jupe(User *u)
{
    char *jserver = strtok(NULL, " ");
    char *reason = strtok(NULL, "");
    char rbuf[256];

    if (!jserver) {
		syntax_error(s_OperServ, u, "JUPE", OPER_JUPE_SYNTAX);
    } else {
    	snprintf(rbuf, sizeof(rbuf), "Juped by %s%s%s", u->nick, reason ? ": " : "", reason ? reason : "");

		send_cmd(NULL, "SQUIT %s :%s", jserver, rbuf);
		send_cmd(NULL, "SERVER %s 2 :%s", jserver, rbuf);
		
		if (WallOSJupe)
			wallops(s_OperServ, "\2%s\2 used JUPE on \2%s\2", u->nick, jserver);
    }
}

/*************************************************************************/

static void do_raw(User *u)
{
    char *text = strtok(NULL, "");

    if (!text)
	syntax_error(s_OperServ, u, "RAW", OPER_RAW_SYNTAX);
    else {
		send_cmd(NULL, "%s", text);
		if (WallOSRaw) {
			char *kw = strtok(text, " ");
    		while (kw && *kw == ':')
    			kw = strtok(NULL, " ");
			wallops(s_OperServ, "\2%s\2 used RAW command for \2%s\2", u->nick, (kw ? kw : "\2non RFC compliant message\2"));
		}
	}
}

/*************************************************************************/

static void do_update(User *u)
{
    notice_lang(s_OperServ, u, OPER_UPDATING);
    save_data = 1;
}

/*************************************************************************/

static void do_reload(User *u)
{
	if (!read_config(1)) {
		quitmsg = calloc(28 + strlen(u->nick), 1);
    	if (!quitmsg)
		quitmsg = "Error during the reload of the configuration file, but out of memory!";
    	else
		sprintf(quitmsg, "Error during the reload of the configuration file!");
    	quitting = 1;
	}
	
	notice_lang(s_OperServ, u, OPER_RELOAD);
}

/*************************************************************************/

static void do_os_quit(User *u)
{
    quitmsg = calloc(28 + strlen(u->nick), 1);
    if (!quitmsg)
	quitmsg = "QUIT command received, but out of memory!";
    else
	sprintf(quitmsg, "QUIT command received from %s", u->nick);
    quitting = 1;
}

/*************************************************************************/

static void do_shutdown(User *u)
{
    quitmsg = calloc(32 + strlen(u->nick), 1);
    if (!quitmsg)
	quitmsg = "SHUTDOWN command received, but out of memory!";
    else
	sprintf(quitmsg, "SHUTDOWN command received from %s", u->nick);
    save_data = 1;
    delayed_quit = 1;
}

/*************************************************************************/

static void do_restart(User *u)
{
#ifdef SERVICES_BIN
    quitmsg = calloc(31 + strlen(u->nick), 1);
    if (!quitmsg)
	quitmsg = "RESTART command received, but out of memory!";
    else
	sprintf(quitmsg, "RESTART command received from %s", u->nick);
    raise(SIGHUP);
#else
    notice_lang(s_OperServ, u, OPER_CANNOT_RESTART);
#endif
}

/*************************************************************************/

/* XXX - this function is broken!! */

static void do_listignore(User *u)
{
    int sent_header = 0;
    IgnoreData *id;
    int i;

    notice(s_OperServ, u->nick, "Command disabled - it's broken.");
    return;
    
    for (i = 0; i < 256; i++) {
	for (id = ignore[i]; id; id = id->next) {
	    if (!sent_header) {
		notice_lang(s_OperServ, u, OPER_IGNORE_LIST);
		sent_header = 1;
	    }
	    notice(s_OperServ, u->nick, "%ld %s", id->time, id->who);
	}
    }
    if (!sent_header)
	notice_lang(s_OperServ, u, OPER_IGNORE_LIST_EMPTY);
}

/*************************************************************************/

#ifdef DEBUG_COMMANDS

static void do_matchwild(User *u)
{
    char *pat = strtok(NULL, " ");
    char *str = strtok(NULL, " ");
    if (pat && str)
	notice(s_OperServ, u->nick, "%d", match_wild(pat, str));
    else
	notice(s_OperServ, u->nick, "Syntax error.");
}

#endif	/* DEBUG_COMMANDS */

/*************************************************************************/

/* Kill all users matching a certain host. The host is obtained from the
 * supplied nick. The raw hostmsk is not supplied with the command in an effort
 * to prevent abuse and mistakes from being made - which might cause *.com to
 * be killed. It also makes it very quick and simple to use - which is usually
 * what you want when someone starts loading numerous clones. In addition to
 * killing the clones, we add a temporary AKILL to prevent them from 
 * immediately reconnecting.
 * Syntax: KILLCLONES nick
 * -TheShadow (29 Mar 1999) 
 */

static void do_killclones(User *u)
{
    char *clonenick = strtok(NULL, " ");
    int count=0;
    User *cloneuser, *user, *tempuser;
    char *clonemask, *akillmask;
    char killreason[NICKMAX+32];
    char akillreason[] = "Temporary KILLCLONES akill.";

    if (!clonenick) {
	notice_lang(s_OperServ, u, OPER_KILLCLONES_SYNTAX);

    } else if (!(cloneuser = finduser(clonenick))) {
	notice_lang(s_OperServ, u, OPER_KILLCLONES_UNKNOWN_NICK, clonenick);

    } else {
	clonemask = scalloc(strlen(GetHost(cloneuser)) + 5, 1);
	sprintf(clonemask, "*!*@%s", GetHost(cloneuser));

	akillmask = scalloc(strlen(GetHost(cloneuser)) + 3, 1);
	sprintf(akillmask, "*@%s", GetHost(cloneuser));
	
	user = firstuser();
	while (user) {
	    if (match_usermask(clonemask, user) != 0) {
		tempuser = nextuser();
		count++;
		snprintf(killreason, sizeof(killreason), 
					"Cloning [%d]", count);
		kill_user(NULL, user->nick, killreason);
		user = tempuser;
	    } else {
		user = nextuser();
	    }
	}

	add_akill(u, akillmask, u->nick, time(NULL) + KillClonesAkillExpire, akillreason);

	wallops(s_OperServ, "\2%s\2 used KILLCLONES for \2%s\2 killing "
			"\2%d\2 clones. A temporary AKILL has been added "
			"for \2%s\2.", u->nick, clonemask, count, akillmask);

	log("%s: KILLCLONES: %d clone(s) matching %s killed.",
			s_OperServ, count, clonemask);

	free(akillmask);
	free(clonemask);
    }
}

/*************************************************************************/
